# Copyright 2016 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Calculations for different cases for additional IP billing
See notes in quark/billing.py for more details.
"""

import click
import datetime

from neutron.common import config
from neutron import context as neutron_context

from pprint import pprint as pp
from quark import billing
from quark.db import models
from quark import network_strategy


def make_case2(context):
    """This is a helper method for testing.

    When run with the current context, it will create a case 2 entries
    in the database. See top of file for what case 2 is.
    """
    query = context.session.query(models.IPAddress)
    period_start, period_end = billing.calc_periods()
    ip_list = billing.build_full_day_ips(query, period_start, period_end)
    import random
    ind = random.randint(0, len(ip_list) - 1)
    address = ip_list[ind]
    address.allocated_at = datetime.datetime.utcnow() -\
        datetime.timedelta(days=1)
    context.session.add(address)
    context.session.flush()


@click.command()
@click.option('--notify', is_flag=True,
              help='If true, sends notifications to billing')
@click.option('--hour', default=0,
              help='period start hour, e.g. 0 is midnight')
@click.option('--minute', default=0,
              help='period start minute, e.g. 0 is top of the hour')
def main(notify, hour, minute):
    """Runs billing report. Optionally sends notifications to billing"""

    # Read the config file and get the admin context
    config_opts = ['--config-file', '/etc/neutron/neutron.conf']
    config.init(config_opts)
    # Have to load the billing module _after_ config is parsed so
    # that we get the right network strategy
    network_strategy.STRATEGY.load()
    billing.PUBLIC_NETWORK_ID = network_strategy.STRATEGY.get_public_net_id()
    config.setup_logging()
    context = neutron_context.get_admin_context()

    # A query to get all IPAddress objects from the db
    query = context.session.query(models.IPAddress)

    (period_start, period_end) = billing.calc_periods(hour, minute)

    full_day_ips = billing.build_full_day_ips(query,
                                              period_start,
                                              period_end)
    partial_day_ips = billing.build_partial_day_ips(query,
                                                    period_start,
                                                    period_end)

    if notify:
        # '==================== Full Day ============================='
        for ipaddress in full_day_ips:
            click.echo('start: {}, end: {}'.format(period_start, period_end))
            payload = billing.build_payload(ipaddress,
                                            billing.IP_EXISTS,
                                            start_time=period_start,
                                            end_time=period_end)
            billing.do_notify(context,
                              billing.IP_EXISTS,
                              payload)
        # '==================== Part Day ============================='
        for ipaddress in partial_day_ips:
            click.echo('start: {}, end: {}'.format(period_start, period_end))
            payload = billing.build_payload(ipaddress,
                                            billing.IP_EXISTS,
                                            start_time=ipaddress.allocated_at,
                                            end_time=period_end)
            billing.do_notify(context,
                              billing.IP_EXISTS,
                              payload)
    else:
        click.echo('Case 1 ({}):\n'.format(len(full_day_ips)))
        for ipaddress in full_day_ips:
            pp(billing.build_payload(ipaddress,
                                     billing.IP_EXISTS,
                                     start_time=period_start,
                                     end_time=period_end))

        click.echo('\n===============================================\n')

        click.echo('Case 2 ({}):\n'.format(len(partial_day_ips)))
        for ipaddress in partial_day_ips:
            pp(billing.build_payload(ipaddress,
                                     billing.IP_EXISTS,
                                     start_time=ipaddress.allocated_at,
                                     end_time=period_end))


if __name__ == '__main__':
    main()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import netaddr

import json

import mock
from oslo_config import cfg

import quark.drivers.ironic_driver
from quark import network_strategy
from quark.tests import test_base


class TestIronicDriverBase(test_base.TestBase):

    def setUp(self):
        super(TestIronicDriverBase, self).setUp()

        net_strategy = {
            "1": {
                "bridge": "publicnet",
                "subnets": {"4": "1", "6": "2"}
            },
            "2": {
                "bridge": "publicnet",
                "subnets": {"3": "1", "6": "4"}
            }
        }
        strategy = json.dumps(net_strategy)
        quark.drivers.ironic_driver.STRATEGY = network_strategy.JSONStrategy(
            strategy)
        cfg.CONF.set_override("default_net_strategy", strategy,
                              "QUARK")

        cfg.CONF.set_override("operation_delay", 0,
                              "IRONIC")
        cfg.CONF.set_override("operation_backoff", 0,
                              "IRONIC")

    @contextlib.contextmanager
    def _stubs(self, create_port=None, delete_port=None):
        importer, client, create, delete = self._create_client(
            create_port, delete_port)
        driver = quark.drivers.ironic_driver.IronicDriver()
        yield driver, client, create, delete

    def _create_client(self, create_port=None, delete_port=None):
        patcher = mock.patch('quark.drivers.ironic_driver.importutils')
        importer_mock = patcher.start()
        self.addCleanup(patcher.stop)

        client_mock = mock.Mock()
        importer_mock.import_class.return_value = client_mock

        create_port_mock = client_mock.return_value.create_port
        if not create_port:
            create_port = [{"port": {"vlan_id": 500, "id": "portid"}}]
        create_port_mock.side_effect = create_port

        delete_port_mock = client_mock.return_value.delete_port
        if delete_port is None:
            delete_port = [None]
        delete_port_mock.side_effect = delete_port
        return importer_mock, client_mock, create_port_mock, delete_port_mock

    def _create_address(self, cidr, default_route=True):
        cidr = netaddr.IPNetwork(cidr)

        gateway_ip = str(netaddr.IPAddress(cidr.first + 1))
        host_ip = str(netaddr.IPAddress(cidr.first + 2))

        dns = [{"ip": "8.8.8.8"}, {"ip": "8.8.4.4"}]

        routes = [{"cidr": "earth", "gateway": gateway_ip},
                  {"cidr": "moon", "gateway": gateway_ip}]

        address = {}
        address["address_readable"] = host_ip

        subnet = {
            "id": "subnet_%s" % str(cidr),
            "name": "subnet_name",
            "tenant_id": "fake",
            "dns_nameservers": dns,
            "routes": routes if not default_route else [],
            "gateway_ip": gateway_ip if default_route else None,
            "cidr": str(cidr)
        }
        address["subnet"] = subnet
        return address

    def _create_base_net_driver(self, driver_type):
        driver = mock.Mock()
        driver.get_lswitch_ids_for_network.return_value = ["lswitch1"]
        driver.get_name.return_value = driver_type
        return driver


class TestIronicDriver(TestIronicDriverBase):

    def test_import_client_class(self):
        importer, _, _, _ = self._create_client()
        quark.drivers.ironic_driver.IronicDriver()
        importer.import_class.assert_called_once_with(
            cfg.CONF.IRONIC.ironic_client)

    def test_client_is_instantiated(self):
        with self._stubs() as (driver, client, _, _):
            params = {
                'endpoint_url': cfg.CONF.IRONIC.endpoint_url,
                'timeout': cfg.CONF.IRONIC.timeout,
                'insecure': cfg.CONF.IRONIC.insecure,
                'ca_cert': cfg.CONF.IRONIC.ca_cert,
                'auth_strategy': cfg.CONF.IRONIC.auth_strategy,
                'tenant_name': cfg.CONF.IRONIC.tenant_name,
                'tenant_id': cfg.CONF.IRONIC.tenant_id,
                'password': cfg.CONF.IRONIC.password
            }
            client.assert_called_once_with(**params)


class TestIronicDriverIPAMStrategies(TestIronicDriverBase):

    def test_ipam_strategies(self):
        with self._stubs() as (driver, client, _, _):
            self.assertEqual(
                driver._ipam_strategies,
                json.loads(cfg.CONF.IRONIC.ironic_ipam_strategies))

    def test_invalid_ipam_strategy_raises(self):
        cfg.CONF.set_override('ironic_ipam_strategies',
                              '{}',
                              'IRONIC')
        self.addCleanup(cfg.CONF.clear_override,
                        'ironic_ipam_strategies',
                        'IRONIC')
        self.assertRaises(quark.drivers.ironic_driver.IronicDriver)

    def test_ipam_strategy_provider_overrides(self):
        with self._stubs() as (driver, _, _, _):
            strategy = driver.select_ipam_strategy("1", "BOTH")
            self.assertEqual(strategy, "IRONIC_BOTH")

    def test_ipam_strategy_provider_returns_default(self):
        with self._stubs() as (driver, _, _, _):
            strategy = driver.select_ipam_strategy("1", "WTF")
            self.assertEqual(strategy, "IRONIC_ANY")

    def test_ipam_strategy_tenant_returns_network_strategy(self):
        with self._stubs() as (driver, _, _, _):
            strategy = driver.select_ipam_strategy("3", "WTF")
            self.assertEqual(strategy, "WTF")


class TestIronicDriverCreatePort(TestIronicDriverBase):

    def test_create_port(self):
        create_response = {
            "port": {
                "id": "port1",
                "vlan_id": 120
            }
        }
        with self._stubs(create_port=[create_response]) as (driver, _,
                                                            create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }
            res = driver.create_port(
                self.context, network_id, port_id,
                **kwargs)

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "UNMANAGED",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': self.context.roles,
                    'dynamic_network': True,
                    'fixed_ips': [],
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(
                res, {"uuid": create_response["port"]["id"],
                      "vlan_id": create_response["port"]["vlan_id"]})
            create.assert_called_once_with(body=expected_call)

    def test_create_port_includes_lswitch_ids(self):
        create_response = {
            "port": {
                "id": "port1",
                "vlan_id": 120
            }
        }
        with self._stubs(create_port=[create_response]) as (driver, _,
                                                            create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "NVP")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }
            res = driver.create_port(
                self.context, network_id, port_id,
                **kwargs)

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "NVP",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': self.context.roles,
                    'lswitch_id': 'lswitch1',
                    'dynamic_network': True,
                    'fixed_ips': [],
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(
                res, {"uuid": create_response["port"]["id"],
                      "vlan_id": create_response["port"]["vlan_id"]})
            create.assert_called_once_with(body=expected_call)

    def test_create_port_includes_roles(self):
        create_response = {
            "port": {
                "id": "port1",
                "vlan_id": 120
            }
        }
        with self._stubs(create_port=[create_response]) as (driver, _,
                                                            create, _):

            self.context.roles = ["role1", "role2"]
            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "NVP")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }
            res = driver.create_port(
                self.context, network_id, port_id,
                **kwargs)

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "NVP",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': ["role1", "role2"],
                    'lswitch_id': 'lswitch1',
                    'dynamic_network': True,
                    'fixed_ips': [],
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(
                res, {"uuid": create_response["port"]["id"],
                      "vlan_id": create_response["port"]["vlan_id"]})
            create.assert_called_once_with(body=expected_call)

    def test_create_port_retries(self):
        create_response = [
            Exception("uhoh"),
            {"port": {"id": "port1", "vlan_id": 120}}
        ]
        with self._stubs(create_port=create_response) as (driver, _,
                                                          create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }
            res = driver.create_port(
                self.context, network_id, port_id,
                **kwargs)

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "UNMANAGED",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': self.context.roles,
                    'dynamic_network': True,
                    'fixed_ips': [],
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(
                res,
                {"uuid": create_response[1]["port"]["id"],
                 "vlan_id": create_response[1]["port"]["vlan_id"]})

            self.assertEqual(
                create.call_args_list,
                [mock.call(body=expected_call),
                 mock.call(body=expected_call)])

    def test_create_port_raises_after_retry_failures(self):
        create_response = [
            Exception("uhoh"),
            Exception("uhoh"),
            Exception("uhoh"),
        ]
        with self._stubs(create_port=create_response) as (driver, _,
                                                          create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }
            self.assertRaises(
                quark.drivers.ironic_driver.IronicException,
                driver.create_port,
                self.context, network_id, port_id, **kwargs)

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "UNMANAGED",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': self.context.roles,
                    'dynamic_network': True,
                    'fixed_ips': [],
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(create.call_count, 3)

            self.assertEqual(
                create.call_args_list,
                [mock.call(body=expected_call),
                 mock.call(body=expected_call),
                 mock.call(body=expected_call)])

    def test_create_port_raises_with_missing_required_kwargs(self):
        with self._stubs() as (driver, _, create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": [],
            }

            for kwarg in ["base_net_driver", "device_id",
                          "instance_node_id", "mac_address"]:

                val = kwargs.pop(kwarg)

                self.assertRaises(
                    quark.drivers.ironic_driver.IronicException,
                    driver.create_port,
                    self.context, network_id, port_id, **kwargs)

                kwargs[kwarg] = val

                create.assert_not_called()

    def test_create_port_raises_with_security_groups(self):
        with self._stubs() as (driver, _, create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [],
                "security_groups": ["foo"],
            }

            self.assertRaises(
                quark.drivers.ironic_driver.IronicException,
                driver.create_port,
                self.context, network_id, port_id, **kwargs)

            create.assert_not_called()

    def test_create_port_with_fixed_ips(self):
        create_response = {
            "port": {
                "id": "port1",
                "vlan_id": 120
            }
        }
        with self._stubs(create_port=[create_response]) as (driver, _,
                                                            create, _):

            network_id = "net1"
            port_id = "port1"

            # mock of the default driver class for the network
            base_net_driver = self._create_base_net_driver(
                "UNMANAGED")

            kwargs = {
                "device_id": "device1",
                "instance_node_id": "nodeid1",
                "mac_address": {"address": 1},
                "base_net_driver": base_net_driver,
                "addresses": [self._create_address("10.0.0.0/30"),
                              self._create_address("10.0.0.5/30",
                                                   default_route=False)],
                "security_groups": [],
            }
            res = driver.create_port(
                self.context, network_id, port_id,
                **kwargs)

            fixed_ips = [
                {'subnet': {
                    'name': 'subnet_name',
                    'tenant_id': 'fake',
                    'dns_nameservers': ['8.8.8.8', '8.8.4.4'],
                    'host_routes': [],
                    'gateway_ip': '10.0.0.1',
                    'cidr': '10.0.0.0/30',
                    'id': 'subnet_10.0.0.0/30'},
                 'ip_address': '10.0.0.2'},
                {'subnet': {
                    'name': 'subnet_name',
                    'tenant_id': 'fake',
                    'dns_nameservers': ['8.8.8.8', '8.8.4.4'],
                    'host_routes': [
                        {'nexthop': '10.0.0.5',
                         'destination': 'earth'},
                        {'nexthop': '10.0.0.5',
                         'destination': 'moon'}
                    ],
                    'gateway_ip': None,
                    'cidr': '10.0.0.5/30',
                    'id': 'subnet_10.0.0.5/30'},
                 'ip_address': '10.0.0.6'}
            ]

            expected_call = {
                'port': {
                    'switch:hardware_id': kwargs["instance_node_id"],
                    'device_owner': '',
                    'network_type': "UNMANAGED",
                    'mac_address': '00:00:00:00:00:01',
                    'network_id': network_id,
                    'tenant_id': self.context.tenant_id,
                    'roles': self.context.roles,
                    'dynamic_network': True,
                    'fixed_ips': fixed_ips,
                    'id': port_id,
                    'device_id': kwargs["device_id"]
                }
            }

            self.assertEqual(
                res, {"uuid": create_response["port"]["id"],
                      "vlan_id": create_response["port"]["vlan_id"]})
            create.assert_called_once_with(body=expected_call)


class TestIronicDriverDeletePort(TestIronicDriverBase):

    def test_delete_port(self):
        with self._stubs(delete_port=[None]) as (driver, client,
                                                 _, delete):
            driver.delete_port(self.context, "foo")
            delete.assert_called_once_with("foo")

    def test_delete_port_retries(self):
        delete_port = [Exception("uhoh"), None]
        with self._stubs(delete_port=delete_port) as (driver,
                                                      client, _,
                                                      delete):
            driver.delete_port(self.context, "foo")
            self.assertEqual(delete.call_count, 2)
            self.assertEqual(
                delete.call_args_list,
                [mock.call("foo"), mock.call("foo")])

    def test_delete_port_fail_does_not_raise(self):
        delete_port = [Exception("uhoh"),
                       Exception("uhoh"),
                       Exception("uhoh")]
        with self._stubs(delete_port=delete_port) as (driver,
                                                      client, _,
                                                      delete):
            driver.delete_port(self.context, "foo")
            self.assertEqual(delete.call_count, 3)
            self.assertEqual(
                delete.call_args_list,
                [mock.call("foo"), mock.call("foo"),
                 mock.call("foo")])

    def test_delete_port_ignores_404(self):
        delete_port = [Exception("404 not found"), None]
        with self._stubs(delete_port=delete_port) as (driver,
                                                      client, _,
                                                      delete):
            driver.delete_port(self.context, "foo")
            delete.assert_called_once_with("foo")


class TestIronicDriverUpdatePort(TestIronicDriverBase):

    def test_update_does_nothing(self):
        with self._stubs() as (driver, client,
                               _, delete):
            res = driver.update_port(self.context, "foo", **{})
            client.update_port.assert_not_called()
            self.assertEqual(res, {"uuid": "foo"})

    def test_update_with_sg_raises(self):
        with self._stubs() as (driver, client,
                               _, delete):
            self.assertRaises(
                quark.drivers.ironic_driver.IronicException,
                driver.update_port,
                self.context, "foo",
                security_groups=["sg1"])


class TestIronicDriverNetwork(TestIronicDriverBase):

    def test_create_network_raises(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.create_network,
                1)

    def test_delete_network_raises(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.delete_network,
                1)

    def test_diag_network_raises(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.diag_network,
                1)


class TestIronicDriverSecurityGroups(TestIronicDriverBase):

    def test_create_security_group(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.create_security_group,
                self.context, "fake")

    def test_delete_security_group(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.delete_security_group,
                self.context, "fake")

    def test_update_security_group(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.update_security_group,
                self.context, "fake")

    def test_create_security_group_rule(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.create_security_group_rule,
                self.context, "fake", "fake_rule")

    def test_delete_security_group_rule(self):
        with self._stubs() as (driver, _, _, _):
            self.assertRaises(
                NotImplementedError,
                driver.delete_security_group_rule,
                self.context, "fake", "fake_rule")



"""Ensure default IP policy exists for subnets without IP policies

Revision ID: 552b213c2b8c
Revises: 45a07fac3d38
Create Date: 2014-07-25 15:07:07.418971

"""

# revision identifiers, used by Alembic.
revision = '552b213c2b8c'
down_revision = '45a07fac3d38'

import logging

from quark.plugin_modules import ip_policies

from alembic import op
from sqlalchemy.sql import column, select, table
from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa

LOG = logging.getLogger("alembic.migration")


def upgrade():
    ip_policy = table('quark_ip_policy',
                      column('id', sa.String(length=36)),
                      column('tenant_id', sa.String(length=255)),
                      column('created_at', sa.DateTime()))
    ip_policy_cidrs = table('quark_ip_policy_cidrs',
                            column('id', sa.String(length=36)),
                            column('created_at', sa.DateTime()),
                            column('ip_policy_id', sa.String(length=36)),
                            column('cidr', sa.String(length=64)))
    subnets = table('quark_subnets',
                    column('id', sa.String(length=36)),
                    column('_cidr', sa.String(length=64)),
                    column('tenant_id', sa.String(length=255)),
                    column('ip_policy_id', sa.String(length=36)))

    connection = op.get_bind()

    # 1. Find all subnets without ip_policy.
    data = connection.execute(select([
        subnets.c.id, subnets.c._cidr, subnets.c.tenant_id]).where(
            subnets.c.ip_policy_id == None)).fetchall()  # noqa
    if not data:
        return

    LOG.info("Subnet IDs without IP policies: %s", [d[0] for d in data])

    # 2. Insert ip_policy rows with id.
    vals = [dict(id=uuidutils.generate_uuid(),
                 created_at=timeutils.utcnow(),
                 tenant_id=tenant_id)
            for id, cidr, tenant_id in data]

    LOG.info("IP Policy IDs to insert: %s", [v["id"] for v in vals])
    connection.execute(ip_policy.insert(), *vals)

    # 3. Insert default ip_policy_cidrs for those ip_policy's.
    vals2 = []
    for ((id, cidr, tenant_id), ip_policy) in zip(data, vals):
        cidrs = []
        ip_policies.ensure_default_policy(cidrs, [dict(cidr=cidr)])
        for cidr in cidrs:
            vals2.append(dict(id=uuidutils.generate_uuid(),
                              created_at=timeutils.utcnow(),
                              ip_policy_id=ip_policy["id"],
                              cidr=str(cidr)))

    LOG.info("IP Policy CIDR IDs to insert: %s", [v["id"] for v in vals2])
    connection.execute(ip_policy_cidrs.insert(), *vals2)

    # 4. Set ip_policy_id rows in quark_subnets.
    for ((id, cidr, tenant_id), ip_policy) in zip(data, vals):
        connection.execute(subnets.update().values(
            ip_policy_id=ip_policy["id"]).where(
                subnets.c.id == id))


def downgrade():
    raise NotImplementedError()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib
import json

import mock
from neutron import context
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import api as db_api
from quark.db import models
from quark import exceptions as q_exc
from quark import network_strategy
from quark import plugin_views
from quark.tests import test_quark_plugin


class TestQuarkGetNetworks(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, nets=None, subnets=None):
        net_mods = []
        subnet_mods = []

        for subnet in subnets:
            subnet_mod = models.Subnet()
            subnet_mod.update(subnet)
            subnet_mods.append(subnet_mod)

        if isinstance(nets, list):
            for net in nets:
                net_mod = models.Network()
                net_mod.update(net)
                net_mod["subnets"] = subnet_mods
                net_mods.append(net_mod)
        else:
            if nets:
                net_mods = nets.copy()
                net_mods["subnets"] = subnet_mods
            else:
                net_mods = nets

        db_mod = "quark.db.api"
        with mock.patch("%s.network_find" % db_mod) as net_find:
            net_find.return_value = net_mods
            yield

    def test_get_networks(self):
        subnet = dict(id=1)
        net = dict(id=1, tenant_id=self.context.tenant_id, name="public",
                   status="ACTIVE")
        with self._stubs(nets=[net], subnets=[subnet]):
            nets = self.plugin.get_networks(self.context, {})
            for key in net.keys():
                self.assertEqual(nets[0][key], net[key])
            self.assertEqual(nets[0]["subnets"][0], 1)

    def test_get_network(self):
        subnet = dict(id=1)
        net = dict(id=1, tenant_id=self.context.tenant_id, name="public",
                   status="ACTIVE")
        expected = net.copy()
        expected["admin_state_up"] = True
        expected["shared"] = False
        expected["status"] = "ACTIVE"
        with self._stubs(nets=net, subnets=[subnet]):
            res = self.plugin.get_network(self.context, 1)
            for key in expected.keys():
                self.assertEqual(res[key], expected[key])
            self.assertEqual(res["subnets"][0], 1)

    def test_get_network_no_network_fails(self):
        with self._stubs(nets=None, subnets=[]):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.get_network(self.context, 1)


class TestQuarkGetNetworksShared(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkGetNetworksShared, self).setUp()
        self.strategy = {"public_network":
                         {"bridge": "xenbr0",
                          "subnets": {"4": "public_v4",
                                      "6": "public_v6"}}}
        self.strategy_json = json.dumps(self.strategy)
        self.old = plugin_views.STRATEGY
        plugin_views.STRATEGY = network_strategy.JSONStrategy(
            self.strategy_json)
        cfg.CONF.set_override("default_net_strategy", self.strategy_json,
                              "QUARK")

    def tearDown(self):
        plugin_views.STRATEGY = self.old

    @contextlib.contextmanager
    def _stubs(self, nets=None, subnets=None):
        net_mods = []

        if isinstance(nets, list):
            for net in nets:
                subnet_mods = []
                subnets = net.pop('subnets', [])

                for subnet in subnets:
                    subnet_mod = models.Subnet()
                    subnet_mod.update(subnet)
                    subnet_mods.append(subnet_mod)

                net_mod = models.Network()
                net_mod.update(net)
                net_mod["subnets"] = subnet_mods
                net_mods.append(net_mod)
        else:
            if nets:
                net_mods = nets.copy()
            else:
                net_mods = nets

        db_mod = "quark.db.api"

        db_api.STRATEGY = network_strategy.JSONStrategy(self.strategy_json)
        network_strategy.STRATEGY = network_strategy.JSONStrategy(
            self.strategy_json)

        with mock.patch("%s._network_find" % db_mod) as net_find:
            net_find.return_value = net_mods
            yield net_find

    def test_get_networks_shared(self):
        net0 = dict(id='public_network', tenant_id=self.context.tenant_id,
                    name="mynet", status="ACTIVE", subnets=[dict(id=0)])
        net1 = dict(id=1, tenant_id=self.context.tenant_id, name="mynet",
                    status="ACTIVE", subnets=[dict(id=1)])
        with self._stubs(nets=[net0, net1]) as net_find:
            ret = self.plugin.get_networks(self.context, None, None, None,
                                           False, {"shared": [True]})
            """ Includes regression for RM8483. """
            for net in ret:
                if net['shared']:
                    self.assertEqual(2, len(net['subnets']))
                else:
                    self.assertEqual(1, len(net['subnets']))
            net_find.assert_called_with(self.context, None, None, None, False,
                                        None, join_subnets=True,
                                        defaults=["public_network"],
                                        provider_query=False)

    def test_get_networks_shared_false(self):
        net0 = dict(id='public_network', tenant_id=self.context.tenant_id,
                    name="mynet", status="ACTIVE", subnets=[dict(id=0)])
        net1 = dict(id=1, tenant_id=self.context.tenant_id, name="mynet",
                    status="ACTIVE")
        with self._stubs(nets=[net0, net1]) as net_find:
            invert = db_api.INVERT_DEFAULTS
            self.plugin.get_networks(self.context, None, None, None, False,
                                     {"shared": [False]})
            net_find.assert_called_with(self.context, None, None, None, False,
                                        None, join_subnets=True,
                                        defaults=[invert, "public_network"],
                                        provider_query=False)

    def test_get_networks_no_shared(self):
        net0 = dict(id='public_network', tenant_id=self.context.tenant_id,
                    name="mynet", status="ACTIVE", subnets=[dict(id=0)])
        net1 = dict(id=1, tenant_id=self.context.tenant_id, name="mynet",
                    status="ACTIVE")
        with self._stubs(nets=[net0, net1]) as net_find:
            self.plugin.get_networks(self.context, None, None, None, False)
            net_find.assert_called_with(self.context, None, None, None,
                                        False, None, join_subnets=True,
                                        defaults=[], provider_query=False)


class TestQuarkGetNetworkCount(test_quark_plugin.TestQuarkPlugin):
    def test_get_port_count(self):
        """This isn't really testable."""
        with mock.patch("quark.db.api.network_count_all"):
            self.plugin.get_networks_count(self.context, {})


class TestQuarkUpdateNetwork(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, net=None):
        net_mod = net
        if net:
            net_mod = net.copy()

        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.network_update" % db_mod)
        ) as (net_find, net_update):
            net_find.return_value = net_mod
            net_update.return_value = net_mod
            yield net_update

    def test_update_network(self):
        net = dict(id=1, tenant_id=self.context.tenant_id)
        new_net = net.copy()
        new_net["ipam_strategy"] = "BOTH_REQUIRED"
        with self._stubs(net=net) as net_update:
            self.plugin.update_network(self.context, 1, dict(network=new_net))
            net_update.assert_called_once_with(
                self.context, net, id=net["id"],
                tenant_id=self.context.tenant_id)

    def test_update_network_not_found_fails(self):
        with self._stubs(net=None):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.update_network(self.context, 1, None)

    def test_update_network_admin_set_ipam_strategy(self):
        net = dict(id=1, tenant_id=self.context.tenant_id)
        new_net = net.copy()
        new_net["ipam_strategy"] = "BOTH_REQUIRED"

        admin_ctx = self.context.elevated()
        with self._stubs(net=net) as net_update:
            self.plugin.update_network(admin_ctx, 1, dict(network=new_net))
            net_update.assert_called_once_with(
                admin_ctx, net, ipam_strategy=new_net["ipam_strategy"],
                id=net["id"], tenant_id=self.context.tenant_id)


class TestQuarkDeleteNetwork(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkDeleteNetwork, self).setUp()
        self.context.is_admin = True

    @contextlib.contextmanager
    def _stubs(self, net=None, ports=None, subnets=None):
        subnets = subnets or []
        net_mod = net
        port_mods = []
        subnet_mods = []

        for port in ports:
            port_model = models.Port()
            port_model.update(port)
            port_mods.append(port_model)

        for subnet in subnets:
            subnet_mod = models.Subnet()
            subnet_mod.update(subnet)
            subnet_mods.append(subnet_mod)

        if net:
            net_mod = models.Network()
            net_mod.update(net)
            net_mod.ports = port_mods
            net_mod["subnets"] = subnet_mods
            net_mod["network_plugin"] = "BASE"

        db_mod = "quark.db.api"
        strategy_prefix = "quark.network_strategy.JSONStrategy"
        with contextlib.nested(
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.network_delete" % db_mod),
            mock.patch("quark.drivers.base.BaseDriver.delete_network"),
            mock.patch("%s.subnet_delete" % db_mod),
            mock.patch("%s.is_provider_network" % strategy_prefix)
        ) as (net_find, net_delete, driver_net_delete, subnet_del,
              is_provider_network):
            net_find.return_value = net_mod
            is_provider_network.return_value = True
            yield net_delete

    def test_delete_network(self):
        net = dict(id=1, tenant_id=self.context.tenant_id)
        with self._stubs(net=net, ports=[]) as net_delete:
            self.plugin.delete_network(self.context, 1)
            self.assertTrue(net_delete.called)

    def test_delete_network_with_ports_fails(self):
        net = dict(id=1, tenant_id=self.context.tenant_id)
        port = dict(id=2)
        with self._stubs(net=net, ports=[port]):
            with self.assertRaises(n_exc.NetworkInUse):
                self.plugin.delete_network(self.context, 1)

    def test_delete_network_not_found_fails(self):
        with self._stubs(net=None, ports=[]):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.delete_network(self.context, 1)

    def test_delete_network_with_subnets_passes(self):
        net = dict(id=1, tenant_id=self.context.tenant_id)
        subnet = dict(id=1)
        with self._stubs(net=net, ports=[], subnets=[subnet]) as net_delete:
            self.plugin.delete_network(self.context, 1)
            self.assertTrue(net_delete.called)

    def test_delete_provider_network_admin(self):
        """Attempts deleting a provider network as an admin tenant"""
        network_id = 'public_network'
        net = dict(id=network_id, tenant_id=self.context.tenant_id)
        subnet = dict(id=1)
        with self._stubs(net=net, ports=[], subnets=[subnet]) as net_delete:
            self.plugin.delete_network(self.context, network_id)
            self.assertTrue(net_delete.called)

    def test_delete_provider_network_non_admin(self):
        """Attempts deleting a provider network as a non-admin tenant"""
        self.context.is_admin = False
        network_id = 'public_network'
        net = dict(id=network_id, tenant_id=self.context.tenant_id)
        subnet = dict(id=1)
        with self._stubs(net=net, ports=[], subnets=[subnet]):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.delete_network(self.context, network_id)


class TestQuarkCreateNetwork(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, net=None, subnet=None, ports=None, find_net=False):
        net_mod = net
        subnet_mod = None
        if net:
            net_mod = models.Network()
            net_mod.update(net)

        if subnet:
            subnet_mod = models.Subnet()
            subnet_mod.update(subnet)

        found_net = None
        if find_net:
            found_net = models.Network()

        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.network_create" % db_mod),
            mock.patch("%s.subnet_create" % db_mod),
            mock.patch("quark.drivers.base.BaseDriver.create_network"),
            mock.patch("%s.network_find" % db_mod)
        ) as (net_create, sub_create, driver_net_create, net_find):
            net_create.return_value = net_mod
            sub_create.return_value = subnet_mod
            net_find.return_value = found_net
            yield net_create

    def test_create_network(self):
        net = dict(id=1, name="public", admin_state_up=True,
                   tenant_id=0, network_plugin="BASE")
        with self._stubs(net=net) as net_create:
            net = self.plugin.create_network(self.context, dict(network=net))
            self.assertTrue(net_create.called)
            self.assertEqual(len(net.keys()), 7)
            self.assertIsNotNone(net["id"])
            self.assertEqual(net["name"], "public")
            self.assertTrue(net["admin_state_up"])
            self.assertEqual(net["status"], "ACTIVE")
            self.assertEqual(net["subnets"], [])
            self.assertEqual(net["shared"], False)
            self.assertEqual(net["tenant_id"], 0)

    def test_create_network_with_subnets(self):
        subnet = dict(id=2, cidr="172.168.0.0/24", tenant_id=0)
        net = dict(id=1, name="public", admin_state_up=True,
                   tenant_id=0)
        with self._stubs(net=net, subnet=subnet) as net_create:
            net.update(dict(subnets=[dict(subnet=subnet)]))
            net = self.plugin.create_network(self.context, dict(network=net))
            self.assertTrue(net_create.called)
            self.assertEqual(len(net.keys()), 7)
            self.assertIsNotNone(net["id"])
            self.assertEqual(net["name"], "public")
            self.assertTrue(net["admin_state_up"])
            self.assertEqual(net["status"], "ACTIVE")
            self.assertEqual(net["subnets"], [2])
            self.assertEqual(net["shared"], False)
            self.assertEqual(net["tenant_id"], 0)

    def test_create_network_with_id(self):
        net = dict(id="abcdef", name="public", admin_state_up=True,
                   tenant_id=0)
        ctxt = context.Context('fake', 'fake', is_admin=True,
                               load_admin_roles=False)
        with self._stubs(net=net):
            res = self.plugin.create_network(ctxt, dict(network=net))
            self.assertEqual(net["id"], res["id"])

    def test_create_network_with_id_already_exists_raises(self):
        net = dict(id="abcdef", name="public", admin_state_up=True,
                   tenant_id=0)
        ctxt = context.Context('fake', 'fake', is_admin=True,
                               load_admin_roles=False)
        with self._stubs(net=net, find_net=True):
            with self.assertRaises(q_exc.NetworkAlreadyExists):
                self.plugin.create_network(ctxt, dict(network=net))

    def test_create_network_with_id_not_admin_ignores_id(self):
        net = dict(id="abcdef", name="public", admin_state_up=True,
                   tenant_id=0)
        with self._stubs(net=net):
            res = self.plugin.create_network(self.context, dict(network=net))
            self.assertNotEqual(net["id"], res["id"])

    def test_create_network_with_ipam_strategy(self):
        net = dict(id="abcdef", name="public", admin_state_up=True,
                   tenant_id=0, ipam_strategy="BOTH")
        admin_context = self.context.elevated()
        original = cfg.CONF.QUARK.show_ipam_strategy

        cfg.CONF.set_override('show_ipam_strategy', True, "QUARK")
        with self._stubs(net=net):
            res = self.plugin.create_network(admin_context, dict(network=net))
            self.assertEqual(res["ipam_strategy"], net["ipam_strategy"])
        cfg.CONF.set_override('show_ipam_strategy', original, "QUARK")

    def test_create_network_with_bad_ipam_strategy_raises(self):
        net = dict(id="abcdef", name="public", admin_state_up=True,
                   tenant_id=0, ipam_strategy="BUSTED")
        admin_context = self.context.elevated()
        with self._stubs(net=net):
            with self.assertRaises(q_exc.InvalidIpamStrategy):
                self.plugin.create_network(admin_context, dict(network=net))


class TestQuarkDiagnoseNetworks(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, nets=None, subnets=None):
        net_mods = []
        subnet_mods = []

        if subnets:
            for subnet in subnets:
                subnet_mod = models.Subnet()
                subnet_mod.update(subnet)
                subnet_mods.append(subnet_mod)

        if nets:
            for net in nets:
                net_mod = models.Network()
                net_mod["network_plugin"] = "BASE"
                net_mod.update(net)
                net_mod["subnets"] = subnet_mods
                net_mods.append(net_mod)
        else:
            if nets:
                net_mods = nets.copy()
                net_mods["subnets"] = subnet_mods
            else:
                net_mods = nets

        db_mod = "quark.db.api"
        with mock.patch("%s.network_find" % db_mod) as net_find:
            net_find.return_value = net_mods
            yield

    def test_diagnose_network_no_network_found(self):
        with self._stubs():
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.diagnose_network(self.context.elevated(), "12345",
                                             None)

    def test_diagnose_network_not_authorized(self):
        with self._stubs():
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.diagnose_network(self.context, "12345",
                                             None)

    def test_diagnose_network_with_wildcard_and_no_networks(self):
        db_mod = "quark.db.api"
        with mock.patch("%s.network_find" % db_mod) as net_find:
            net_find.return_value = []
            actual = self.plugin.diagnose_network(self.context.elevated(),
                                                  "*", {})
            expected = {'networks': []}
            self.assertEqual(expected, actual)

    def test_diagnose_network_with_wildcard_and_networks(self):
        subnet = dict(id=1)
        net = dict(id=1, tenant_id=self.context.tenant_id, name="public",
                   status="ACTIVE", network_plugin="BASE")
        with self._stubs(nets=[net], subnets=[subnet]):
            db_mod = "quark.db.api"
            with mock.patch("%s.network_find" % db_mod) as net_find:
                net_find.return_value = [net]
                nets = self.plugin.diagnose_network(self.context.elevated(),
                                                    "*", {})
                net.pop("network_plugin")
                for key in net.keys():
                    self.assertEqual(nets['networks'][0][key], net[key])



"""add_segment_allocations

Revision ID: 1bd7cff90384
Revises: 374c1bdb4480
Create Date: 2016-01-25 19:26:27.899610

"""

# revision identifiers, used by Alembic.
revision = '1bd7cff90384'
down_revision = '374c1bdb4480'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'quark_segment_allocation_ranges',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('segment_id', sa.String(length=36), nullable=True),
        sa.Column('segment_type', sa.String(length=36), nullable=True),
        sa.Column('first_id', sa.BigInteger(), nullable=False),
        sa.Column('last_id', sa.BigInteger(), nullable=False),
        sa.Column('do_not_use', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_quark_segment_allocation_ranges_segment_id'),
                    'quark_segment_allocation_ranges', ['segment_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_segment_allocation_ranges_segment_type'),
                    'quark_segment_allocation_ranges', ['segment_type'],
                    unique=False)
    op.create_table(
        'quark_segment_allocations',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.BigInteger(), autoincrement=False,
                  nullable=False),
        sa.Column('segment_id', sa.String(length=36), nullable=False),
        sa.Column('segment_type', sa.String(length=36), nullable=False),
        sa.Column('segment_allocation_range_id', sa.String(length=36),
                  nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('deallocated', sa.Boolean(), nullable=True),
        sa.Column('deallocated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['segment_allocation_range_id'],
                                ['quark_segment_allocation_ranges.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', 'segment_id', 'segment_type'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_quark_segment_allocations_deallocated'),
                    'quark_segment_allocations', ['deallocated'],
                    unique=False)
    op.create_index(op.f('ix_quark_segment_allocations_deallocated_at'),
                    'quark_segment_allocations', ['deallocated_at'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_segment_allocations_deallocated_at'),
                  table_name='quark_segment_allocations')
    op.drop_index(op.f('ix_quark_segment_allocations_deallocated'),
                  table_name='quark_segment_allocations')
    op.drop_table('quark_segment_allocations')
    op.drop_index(op.f('ix_quark_segment_allocation_ranges_segment_type'),
                  table_name='quark_segment_allocation_ranges')
    op.drop_index(op.f('ix_quark_segment_allocation_ranges_segment_id'),
                  table_name='quark_segment_allocation_ranges')
    op.drop_table('quark_segment_allocation_ranges')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
import webob

from quark.db import api as db_api
from quark.db import ip_types
from quark import exceptions as q_exc
from quark import ipam
from quark import plugin_views as v

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

# NOTE(roaet): this number includes the assumed 'given ips' and any additional
total_ips_on_port = {'00000000-0000-0000-0000-000000000000': 6,
                     '11111111-1111-1111-1111-111111111111': 1,
                     '*': 5}

# NOTE(roaet): shared ips are treated like any other IP and will consume
# from total_ips_on_port
shared_ip_on_network = {'00000000-0000-0000-0000-000000000000': 5,
                        '11111111-1111-1111-1111-111111111111': 0,
                        '*': -1}

quark_ip_addr_opts = [
    cfg.BoolOpt('ipaddr_allow_fixed_ip',
                default=False,
                help=_('Controls if /ip_addresses can make fixed IPs or not')),
    cfg.DictOpt('total_ips_allowed_on_port',
                default=total_ips_on_port,
                help=_('Defines how many IPs may be on a port by network in '
                       'key:quantity format')),
    cfg.DictOpt('shared_ips_allowed_on_network',
                default=shared_ip_on_network,
                help=_('Defines how many shared IPs may be made on a network '
                       'in key:quantity format'))
]

CONF.register_opts(quark_ip_addr_opts, "QUARK")

# NOTE(thomasem): Since IP addresses are only at the subnet level, use
# the QuarkIpamANY strategy, due to other IPAM strategies only being
# relevant at the network level (port create).
ipam_driver = ipam.IPAM_REGISTRY.get_strategy(ipam.QuarkIpamANY.get_name())


def _can_add_ip_to_port_on_network(context, network_id, db_port):
    return True


def _can_create_shared_ip_on_network(context, network_id):
    return True


def get_ip_addresses(context, **filters):
    LOG.info("get_ip_addresses for tenant %s" % context.tenant_id)
    if not filters:
        filters = {}
    if 'type' in filters:
        filters['address_type'] = filters['type']
    filters["_deallocated"] = False
    addrs = db_api.ip_address_find(context, scope=db_api.ALL, **filters)
    return [v._make_ip_dict(ip) for ip in addrs]


def get_ip_address(context, id):
    LOG.info("get_ip_address %s for tenant %s" %
             (id, context.tenant_id))
    filters = {}
    filters["_deallocated"] = False
    addr = db_api.ip_address_find(context, id=id, scope=db_api.ONE, **filters)
    if not addr:
        raise q_exc.IpAddressNotFound(addr_id=id)
    return v._make_ip_dict(addr)


def validate_and_fetch_segment(ports, network_id):
    first_segment = None
    segment_id = None
    for port in ports:
        addresses = port.get("ip_addresses", [])
        for address in addresses:
            if address["network_id"] != network_id:
                raise n_exc.BadRequest(resource="ip_addresses",
                                       msg="Must have ports connected to"
                                           " the requested network")
            segment_id = address.subnet.get("segment_id")
            first_segment = first_segment or segment_id
            if segment_id != first_segment:
                raise n_exc.BadRequest(resource="ip_addresses",
                                       msg="Segment id's do not match.")
    return segment_id


def validate_port_ip_quotas(context, network, ports):
    if network not in CONF.QUARK.total_ips_allowed_on_port:
        network = '*'
    limit = CONF.QUARK.total_ips_allowed_on_port.get(network)
    if limit < 0:
        return
    for port in ports:
        if len(port.associations) + 1 > limit:
            raise q_exc.CannotAddMoreIPsToPort()


def validate_shared_ips_quotas(context, network, addresses):
    if network not in CONF.QUARK.shared_ips_allowed_on_network:
        network = '*'
    limit = CONF.QUARK.shared_ips_allowed_on_network.get(network)
    if limit < 0:
        return
    if len(addresses) + 1 > limit:
        raise q_exc.CannotCreateMoreSharedIPs()


def _shared_ip_request(ip_address):
    port_ids = ip_address.get('ip_address', {}).get('port_ids', [])
    device_ids = ip_address.get('ip_address', {}).get('device_ids', [])
    return len(port_ids) > 1 or len(device_ids) > 1


def _shared_ip_and_active(ip_address, except_port=None):
    if ip_address.is_shared() and ip_address.has_shared_owner():
        return True
    return False


def _can_be_shared(address_model):
    # Don't share IP if any of the assocs is enabled
    return not any(a.enabled for a in address_model.associations)


def create_ip_address(context, body):
    LOG.info("create_ip_address for tenant %s" % context.tenant_id)
    iptype = (ip_types.SHARED if _shared_ip_request(body)
              else ip_types.FIXED)
    if 'ip_address' not in body:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="Invalid request body.")
    if iptype == ip_types.FIXED and not CONF.QUARK.ipaddr_allow_fixed_ip:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="Only shared IPs may be made with "
                                   "this resource.")
    ip_dict = body.get("ip_address")
    port_ids = ip_dict.get('port_ids', [])
    network_id = ip_dict.get('network_id')
    tenant_id = ip_dict.get('tenant_id')
    device_ids = ip_dict.get('device_ids')
    ip_version = ip_dict.get('version')
    ip_address = ip_dict.get('ip_address')
    # If no version is passed, you would get what the network provides,
    # which could be both v4 and v6 addresses. Rather than allow for such
    # an ambiguous outcome, we'll raise instead
    if not ip_version:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="version is required.")
    if network_id is None:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="network_id is required.")
    if network_id == "":
        raise n_exc.NetworkNotFound(net_id=network_id)
    net = db_api.network_find(context, None, None, None, False,
                              id=network_id, scope=db_api.ONE)
    if not net:
        raise n_exc.NetworkNotFound(net_id=network_id)
    if not port_ids and not device_ids:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="port_ids or device_ids required.")

    new_addresses = []
    ports = []
    by_device = False
    with context.session.begin():
        if network_id and device_ids:
            by_device = True
            for device_id in device_ids:
                port = db_api.port_find(
                    context, network_id=network_id, device_id=device_id,
                    tenant_id=tenant_id, scope=db_api.ONE)
                if port is not None:
                    ports.append(port)
        elif port_ids:
            for port_id in port_ids:

                port = db_api.port_find(context, id=port_id,
                                        tenant_id=tenant_id,
                                        scope=db_api.ONE)
                if port is not None:
                    ports.append(port)

        if not ports:
            raise n_exc.PortNotFoundOnNetwork(port_id=port_ids,
                                              net_id=network_id)

    if ((by_device and len(device_ids) != len(ports)) or
            (not by_device and len(port_ids) != len(ports))):
        raise q_exc.NotAllPortOrDeviceFound()

    segment_id = validate_and_fetch_segment(ports, network_id)
    if iptype == ip_types.SHARED:
        old_addresses = db_api.ip_address_find(context,
                                               network_id=network_id,
                                               tenant_id=tenant_id,
                                               address_type=ip_types.SHARED,
                                               scope=db_api.ALL)
        validate_shared_ips_quotas(context, network_id, old_addresses)
    validate_port_ip_quotas(context, network_id, ports)

    # Shared Ips are only new IPs. Two use cases: if we got device_id
    # or if we got port_ids. We should check the case where we got port_ids
    # and device_ids. The device_id must have a port on the network,
    # and any port_ids must also be on that network already. If we have
    # more than one port by this step, it's considered a shared IP,
    # and therefore will be marked as unconfigured (enabled=False)
    # for all ports.
    ipam_driver.allocate_ip_address(context, new_addresses, network_id,
                                    None, CONF.QUARK.ipam_reuse_after,
                                    version=ip_version,
                                    ip_addresses=[ip_address]
                                    if ip_address else [],
                                    segment_id=segment_id,
                                    address_type=iptype)
    with context.session.begin():
        address = new_addresses[0]
        new_address = db_api.port_associate_ip(context, ports, address)
    return v._make_ip_dict(new_address)


def _get_deallocated_override():
    """This function exists to mock and for future requirements if needed."""
    return '2000-01-01 00:00:00'


def _raise_if_shared_and_enabled(address_request, address_model):
    if (_shared_ip_request(address_request)
            and not _can_be_shared(address_model)):
        raise n_exc.BadRequest(
            resource="ip_addresses",
            msg="This IP address is in use on another port and cannot be "
                "shared")


def update_ip_address(context, id, ip_address):
    """Due to NCP-1592 ensure that address_type cannot change after update."""
    LOG.info("update_ip_address %s for tenant %s" % (id, context.tenant_id))
    ports = []
    if 'ip_address' not in ip_address:
        raise n_exc.BadRequest(resource="ip_addresses",
                               msg="Invalid request body.")
    with context.session.begin():
        address = db_api.ip_address_find(context, id=id, scope=db_api.ONE)
        if not address:
            raise q_exc.IpAddressNotFound(addr_id=id)
        iptype = address.address_type
        if iptype == ip_types.FIXED and not CONF.QUARK.ipaddr_allow_fixed_ip:
            raise n_exc.BadRequest(
                resource="ip_addresses",
                msg="Fixed ips cannot be updated using this interface.")

        reset = ip_address['ip_address'].get('reset_allocation_time', False)
        if reset and address['deallocated'] == 1:
            if context.is_admin:
                LOG.info("IP's deallocated time being manually reset")
                address['deallocated_at'] = _get_deallocated_override()
            else:
                msg = "Modification of reset_allocation_time requires admin"
                raise webob.exc.HTTPForbidden(detail=msg)

        port_ids = ip_address['ip_address'].get('port_ids', None)

        if port_ids is not None and not port_ids:
            raise n_exc.BadRequest(
                resource="ip_addresses",
                msg="Cannot be updated with empty port_id list")

        if iptype == ip_types.SHARED:
            has_owner = address.has_any_shared_owner()

        if port_ids:
            if iptype == ip_types.FIXED and len(port_ids) > 1:
                raise n_exc.BadRequest(
                    resource="ip_addresses",
                    msg="Fixed ips cannot be updated with more than one port.")

            _raise_if_shared_and_enabled(ip_address, address)
            ports = db_api.port_find(context, tenant_id=context.tenant_id,
                                     id=port_ids, scope=db_api.ALL)
            # NOTE(name): could be considered inefficient because we're
            # converting to a list to check length. Maybe revisit
            if len(ports) != len(port_ids):
                raise n_exc.PortNotFound(port_id=port_ids)

            validate_and_fetch_segment(ports, address["network_id"])
            validate_port_ip_quotas(context, address.network_id, ports)

            if iptype == ip_types.SHARED and has_owner:
                for assoc in address.associations:
                    pid = assoc.port_id
                    if pid not in port_ids and 'none' != assoc.service:
                        raise q_exc.PortRequiresDisassociation()

            LOG.info("Updating IP address, %s, to only be used by the"
                     "following ports:  %s" % (address.address_readable,
                                               [p.id for p in ports]))
            new_address = db_api.update_port_associations_for_ip(context,
                                                                 ports,
                                                                 address)
        elif iptype == ip_types.SHARED and has_owner:
            raise q_exc.PortRequiresDisassociation()
        else:
            ipam_driver.deallocate_ip_address(context, address)
            return v._make_ip_dict(address)
    return v._make_ip_dict(new_address)


def delete_ip_address(context, id):
    """Delete an ip address.

    : param context: neutron api request context
    : param id: UUID representing the ip address to delete.
    """
    LOG.info("delete_ip_address %s for tenant %s" % (id, context.tenant_id))
    with context.session.begin():
        ip_address = db_api.ip_address_find(
            context, id=id, scope=db_api.ONE)
        if not ip_address or ip_address.deallocated:
            raise q_exc.IpAddressNotFound(addr_id=id)

        iptype = ip_address.address_type
        if iptype == ip_types.FIXED and not CONF.QUARK.ipaddr_allow_fixed_ip:
            raise n_exc.BadRequest(
                resource="ip_addresses",
                msg="Fixed ips cannot be updated using this interface.")

        if ip_address.has_any_shared_owner():
            raise q_exc.PortRequiresDisassociation()

        db_api.update_port_associations_for_ip(context, [], ip_address)

        ipam_driver.deallocate_ip_address(context, ip_address)


def get_ports_for_ip_address(context, ip_id, limit=None, sorts=None,
                             marker=None, page_reverse=False, filters=None,
                             fields=None):
    """Retrieve a list of ports.

    The contents of the list depends on the identity of the user
    making the request (as indicated by the context) as well as any
    filters.
    : param context: neutron api request context
    : param filters: a dictionary with keys that are valid keys for
        a port as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.
    : param fields: a list of strings that are valid keys in a
        port dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_ports for tenant %s filters %s fields %s" %
             (context.tenant_id, filters, fields))
    addr = db_api.ip_address_find(context, id=ip_id, scope=db_api.ONE)
    if not addr:
        raise q_exc.IpAddressNotFound(addr_id=ip_id)

    if filters is None:
        filters = {}

    filters['ip_address_id'] = [ip_id]

    ports = db_api.port_find(context, limit, sorts, marker,
                             fields=fields, join_security_groups=True,
                             **filters)
    return v._make_ip_ports_list(addr, ports, fields)


def get_port_for_ip_address(context, ip_id, id, fields=None):
    """Retrieve a port.

    : param context: neutron api request context
    : param id: UUID representing the port to fetch.
    : param fields: a list of strings that are valid keys in a
        port dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_port %s for tenant %s fields %s" %
             (id, context.tenant_id, fields))
    addr = db_api.ip_address_find(context, id=ip_id, scope=db_api.ONE)
    if not addr:
        raise q_exc.IpAddressNotFound(addr_id=ip_id)

    filters = {'ip_address_id': [ip_id]}
    results = db_api.port_find(context, id=id, fields=fields,
                               scope=db_api.ONE, **filters)

    if not results:
        raise n_exc.PortNotFound(port_id=id)

    return v._make_port_for_ip_dict(addr, results)


def update_port_for_ip_address(context, ip_id, id, port):
    """Update values of a port.

    : param context: neutron api request context
    : param ip_id: UUID representing the ip associated with port to update
    : param id: UUID representing the port to update.
    : param port: dictionary with keys indicating fields to update.
        valid keys are those that have a value of True for 'allow_put'
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.
    """
    LOG.info("update_port %s for tenant %s" % (id, context.tenant_id))
    sanitize_list = ['service']
    with context.session.begin():
        addr = db_api.ip_address_find(context, id=ip_id, scope=db_api.ONE)
        if not addr:
            raise q_exc.IpAddressNotFound(addr_id=ip_id)
        port_db = db_api.port_find(context, id=id, scope=db_api.ONE)
        if not port_db:
            raise q_exc.PortNotFound(port_id=id)
        port_dict = {k: port['port'][k] for k in sanitize_list}

        require_da = False
        service = port_dict.get('service')

        if require_da and _shared_ip_and_active(addr, except_port=id):
            raise q_exc.PortRequiresDisassociation()
        addr.set_service_for_port(port_db, service)
        context.session.add(addr)
    return v._make_port_for_ip_dict(addr, port_db)



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neutron.api import extensions

RESOURCE_NAME = "network"
RESOURCE_COLLECTION = RESOURCE_NAME + "s"
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {
        "ipam_strategy": {"allow_post": True, "allow_put": False,
                          "is_visible": True, "default": False},
        "network_plugin": {"allow_post": True, "allow_put": False,
                           "is_visible": False, "default": False},
        "id": {"allow_post": True, "allow_put": False,
               "is_visible": True, "default": False}}}


class Networks_quark(extensions.ExtensionDescriptor):
    """Extends Networks for quark API purposes."""

    @classmethod
    def get_name(cls):
        return "Quark Networks API Extension"

    @classmethod
    def get_alias(cls):
        return "networks_quark"

    @classmethod
    def get_description(cls):
        return "Quark Networks API Extension"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "networks_quark/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2013-03-25T19:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.extensions import securitygroup as sg_ext
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark import exceptions as q_exc


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

# Neutron doesn't officially support any other ethertype
ETHERTYPES = {
    "IPv4": 0x0800,
    "IPv6": 0x86DD
}

PROTOCOLS_V4 = {"icmp": 1, "tcp": 6, "udp": 17}
PROTOCOLS_V6 = {"tcp": 6, "udp": 17, "icmp": 58}

# Neutron only officially supports TCP, ICMP and UDP,
# with ethertypes IPv4 and IPv6
PROTOCOL_MAP = {
    ETHERTYPES["IPv4"]: PROTOCOLS_V4,
    ETHERTYPES["IPv6"]: PROTOCOLS_V6
}


ALLOWED_PROTOCOLS = None
ALLOWED_WITH_RANGE = [1, 6, 17, 58]
MIN_PROTOCOL = 0
MAX_PROTOCOL = 255
REVERSE_PROTOCOL_MAP = {}
REVERSE_ETHERTYPES = {}
MIN_PORT = 0
MAX_PORT = 65535


def _is_allowed(protocol, ethertype):
    # Please see http://en.wikipedia.org/wiki/List_of_IP_protocol_numbers
    # The field is always 8 bits wide.
    if not (MIN_PROTOCOL <= protocol <= MAX_PROTOCOL):
        return False

    return (protocol in PROTOCOL_MAP[ethertype] or
            protocol in REVERSE_PROTOCOL_MAP)


def translate_ethertype(ethertype):
    if ethertype not in ETHERTYPES:
        raise q_exc.InvalidEthertype(ethertype=ethertype)
    return ETHERTYPES[ethertype]


def translate_protocol(protocol, ethertype):
    ether = translate_ethertype(ethertype)
    try:
        proto = int(protocol)
    except ValueError:
        proto = str(protocol).lower()
        proto = PROTOCOL_MAP[ether].get(proto, -1)

    if not _is_allowed(proto, ether):
        # TODO(mdietz) This will change as neutron supports new protocols
        value_list = PROTOCOL_MAP[ETHERTYPES["IPv4"]].keys()
        raise sg_ext.SecurityGroupRuleInvalidProtocol(
            protocol=protocol, values=value_list)
    return proto


def human_readable_ethertype(ethertype):
    return REVERSE_ETHERTYPES[ethertype]


def human_readable_protocol(protocol, ethertype):
    if protocol is None:
        return
    proto = translate_protocol(protocol, ethertype)
    return REVERSE_PROTOCOL_MAP[proto]


def validate_remote_ip_prefix(ethertype, prefix):
    if prefix:
        net = netaddr.IPNetwork(prefix)
        if ((ethertype == ETHERTYPES["IPv4"] and net.version == 6) or
                (ethertype == ETHERTYPES["IPv6"] and net.version == 4)):
            human_ether = human_readable_ethertype(ethertype)
            raise n_exc.InvalidInput(
                error_message="Etherytype %s does not match "
                              "remote_ip_prefix, which is IP version %s" %
                              (human_ether, net.version))


def validate_protocol_with_port_ranges(ethertype, protocol, port_range_min,
                                       port_range_max):
    if protocol in ALLOWED_WITH_RANGE:
        if protocol == PROTOCOL_MAP[ethertype]["icmp"]:
            if port_range_min is None and port_range_max is not None:
                raise sg_ext.SecurityGroupMissingIcmpType(value=port_range_max)
            elif port_range_min is not None:
                attr = None
                field = None
                value = None
                if port_range_min < 0 or port_range_min > 255:
                    field = "port_range_min"
                    attr = "type"
                    value = port_range_min
                elif (port_range_max is not None and
                      port_range_max < 0 or port_range_max > 255):
                    field = "port_range_max"
                    attr = "code"
                    value = port_range_max

                if attr and field and value:
                    raise sg_ext.SecurityGroupInvalidIcmpValue(
                        field=field, attr=attr, value=value)

        else:
            if (port_range_min is None) != (port_range_max is None):
                # TODO(anyone): what exactly is a TCP or UDP rule withouts
                #               ports?
                raise n_exc.InvalidInput(
                    error_message="For TCP/UDP rules, port_range_min and"
                                  "port_range_max must either both be supplied"
                                  ", or neither of them")

            if port_range_min is not None and port_range_max is not None:
                if port_range_min > port_range_max:
                    raise sg_ext.SecurityGroupInvalidPortRange()

                if port_range_min < MIN_PORT or port_range_max > MAX_PORT:
                    raise n_exc.InvalidInput(
                        error_message="port_range_min and port_range_max must "
                                      "be >= %s and <= %s" % (MIN_PORT,
                                                              MAX_PORT))


def _init_protocols():
    if not REVERSE_PROTOCOL_MAP:
        for ether_str, ethertype in ETHERTYPES.iteritems():
            for proto, proto_int in PROTOCOL_MAP[ethertype].iteritems():
                REVERSE_PROTOCOL_MAP[proto_int] = proto.upper()

    if not REVERSE_ETHERTYPES:
        for ether_str, ethertype in ETHERTYPES.iteritems():
            REVERSE_ETHERTYPES[ethertype] = ether_str


_init_protocols()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.


import mock
from oslo_config import cfg

import quark.plugin
from quark.tests.functional.base import BaseFunctionalTest


class TestQuarkPlugin(BaseFunctionalTest):
    def setUp(self):
        super(TestQuarkPlugin, self).setUp()

        cfg.CONF.set_override('quota_ports_per_network', 1, 'QUOTAS')
        self.plugin = quark.plugin.Plugin()


class TestQuarkAPIExtensions(TestQuarkPlugin):
    """Adds coverage for appending the API extension path."""
    def test_append_quark_extensions(self):
        conf = mock.MagicMock()
        conf.__contains__.return_value = False
        quark.plugin.append_quark_extensions(conf)
        self.assertEqual(conf.set_override.call_count, 0)

    def test_append_no_extension_path(self):
        conf = mock.MagicMock()
        conf.__contains__.return_value = True
        with mock.patch("quark.plugin.extensions") as extensions:
            extensions.__path__ = ["apple", "banana", "carrot"]
            quark.plugin.append_quark_extensions(conf)
            conf.__contains__.assert_called_once_with("api_extensions_path")
            conf.set_override.assert_called_once_with(
                "api_extensions_path",
                "apple:banana:carrot")



"""IP Address reallocation optimizations

Revision ID: e249ebc4f51
Revises: 5932938bb839
Create Date: 2015-08-20 01:08:04.554914

"""

# revision identifiers, used by Alembic.
revision = 'e249ebc4f51'
down_revision = '5932938bb839'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_reallocate_ip_addresses_network'),
                    'quark_ip_addresses',
                    ["network_id", "deallocated_at", "version",
                     "_deallocated"],
                    unique=False)
    op.create_index(op.f('ix_quark_reallocate_ip_addresses_subnet'),
                    'quark_ip_addresses',
                    ["deallocated_at", "subnet_id", "_deallocated"],
                    unique=False)


def downgrade():
    # These will never be run, upstream has disabled rollbacks
    op.drop_index(op.f('ix_quark_reallocate_ip_addresses_network'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_reallocate_ip_addresses_subnet'),
                  table_name='quark_ip_addresses')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Ironic agent driver for Quark.
"""
import json
import netaddr

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from quark.drivers import base
from quark import network_strategy
from quark import utils

STRATEGY = network_strategy.STRATEGY

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

IRONIC_IPAM_STRATEGIES = {
    "provider": {
        "default": "IRONIC_ANY",
        "overrides": {
            "BOTH_REQUIRED": "IRONIC_BOTH_REQUIRED",
            "BOTH": "IRONIC_BOTH",
            "ANY": "IRONIC_ANY"
        }
    },
    "tenant": {
        "overrides": {}
    }
}

ironic_opts = [
    # client connection info
    cfg.StrOpt("ironic_client", default="neutronclient.v2_0.client.Client",
               help=("Class of the client used to communicate with the "
                     "Ironic agent")),
    cfg.StrOpt('endpoint_url',
               default='http://127.0.0.1:9697',
               help='URL for connecting to neutron'),
    cfg.IntOpt('timeout',
               default=30,
               help='Timeout value for connecting to neutron in seconds.'),
    cfg.BoolOpt('insecure',
                default=False,
                help='If set, ignore any SSL validation issues'),
    cfg.StrOpt('ca_cert',
               help=('Location of CA certificates file to use for '
                     'neutron client requests.')),
    cfg.StrOpt('password',
               help='Password for connecting to neutron.', secret=True),
    cfg.StrOpt('tenant_name',
               help='Tenant name for connecting to neutron.'),
    cfg.StrOpt('tenant_id',
               help='Tenant id for connecting to neutron.'),
    cfg.StrOpt('auth_url',
               default='http://localhost:5000/v2.0',
               help='Authorization URL for connecting to neutron.'),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               help='Authorization strategy for connecting to neutron.'),

    # client retry options
    cfg.IntOpt("operation_retries",
               default=3,
               help="Number of times to attempt client operations."),
    cfg.IntOpt("operation_delay",
               default=0,
               help="Base seconds to wait between client attempts."),
    cfg.IntOpt("operation_backoff",
               default=0,
               help="Base seconds for client exponential backoff"),

    # conf options
    cfg.StrOpt("ironic_ipam_strategies",
               default=json.dumps(IRONIC_IPAM_STRATEGIES),
               help="Default IPAM strategies and overrides for this driver")
]

CONF.register_opts(ironic_opts, "IRONIC")


class FakeIronicClient(object):
    def __init__(self, *args, **kwargs):
        pass

    def create_port(self, *args, **kwargs):
        return {"port": {"vlan_id": 500, "id": "fake_uuid"}}

    def delete_port(self, *args, **kwargs):
        return


class IronicException(n_exc.NeutronException):
    message = "ironic driver error: %(msg)s"


class IronicDriver(base.BaseDriver):
    def __init__(self):
        self._client = None
        self._ipam_strategies = None
        super(IronicDriver, self).__init__()

    @classmethod
    def get_name(cls):
        return "IRONIC"

    def load_config(self):
        LOG.info("Loading Ironic settings.")
        self._client_cls = importutils.import_class(CONF.IRONIC.ironic_client)
        self._client = self._get_client()
        self._ipam_strategies = self._parse_ironic_ipam_strategies()
        LOG.info("Ironic Driver config loaded. Client: %s"
                 % (CONF.IRONIC.endpoint_url))

    def _get_client(self):

        params = {
            'endpoint_url': CONF.IRONIC.endpoint_url,
            'timeout': CONF.IRONIC.timeout,
            'insecure': CONF.IRONIC.insecure,
            'ca_cert': CONF.IRONIC.ca_cert,
            'auth_strategy': CONF.IRONIC.auth_strategy,
            'tenant_name': CONF.IRONIC.tenant_name,
            'tenant_id': CONF.IRONIC.tenant_id,
            'password': CONF.IRONIC.password
        }
        return self._client_cls(**params)

    def _parse_ironic_ipam_strategies(self):
        strategies = json.loads(CONF.IRONIC.ironic_ipam_strategies)

        for net_type in ['provider', 'tenant']:

            strategy = strategies.get(net_type, {})
            default = strategy.get("default")

            # provider nets must specify a default IPAM strategy
            if net_type == 'provider':
                if not default:
                    raise Exception("ironic_ipam_strategies must have a "
                                    "default 'provider' strategy.")

        return strategies

    def select_ipam_strategy(self, network_id, network_strategy, **kwargs):
        """Return relevant IPAM strategy name.

        :param network_id: neutron network id.
        :param network_strategy: default strategy for the network.

        NOTE(morgabra) This feels like a hack but I can't think of a better
        idea. The root problem is we can now attach ports to networks with
        a different backend driver/ipam strategy than the network speficies.

        We handle the the backend driver part with allowing network_plugin to
        be specified for port objects. This works pretty well because nova or
        whatever knows when we are hooking up an Ironic node so it can pass
        along that key during port_create().

        IPAM is a little trickier, especially in Ironic's case, because we
        *must* use a specific IPAM for provider networks. There isn't really
        much of an option other than involve the backend driver when selecting
        the IPAM strategy.
        """
        LOG.info("Selecting IPAM strategy for network_id:%s "
                 "network_strategy:%s" % (network_id, network_strategy))

        net_type = "tenant"
        if STRATEGY.is_provider_network(network_id):
            net_type = "provider"

        strategy = self._ipam_strategies.get(net_type, {})
        default = strategy.get("default")
        overrides = strategy.get("overrides", {})

        # If we override a particular strategy explicitly, we use it.
        if network_strategy in overrides:
            LOG.info("Selected overridden IPAM strategy: %s"
                     % (overrides[network_strategy]))
            return overrides[network_strategy]

        # Otherwise, we are free to use an explicit default.
        if default:
            LOG.info("Selected default IPAM strategy for tenant "
                     "network: %s" % (default))
            return default

        # Fallback to the network-specified IPAM strategy
        LOG.info("Selected network strategy for tenant "
                 "network: %s" % (network_strategy))
        return network_strategy

    def _make_subnet_dict(self, subnet):

        dns_nameservers = [str(netaddr.IPAddress(dns["ip"]))
                           for dns in subnet.get("dns_nameservers")]

        host_routes = [{"destination": r["cidr"],
                        "nexthop": r["gateway"]} for r in subnet["routes"]]

        res = {
            "id": subnet.get("id"),
            "name": subnet.get("name"),
            "tenant_id": subnet.get("tenant_id"),
            "dns_nameservers": dns_nameservers,
            "host_routes": host_routes,
            "cidr": subnet.get("cidr"),
            "gateway_ip": subnet.get("gateway_ip")
        }

        return res

    def _make_fixed_ip_dict(self, context, address):
        return {"subnet": self._make_subnet_dict(address["subnet"]),
                "ip_address": address["address_readable"]}

    @utils.retry_loop(CONF.IRONIC.operation_retries,
                      delay=CONF.IRONIC.operation_delay,
                      backoff=CONF.IRONIC.operation_backoff)
    def _create_port(self, context, body):
        try:
            return self._client.create_port(body={"port": body})
        except Exception as e:
            msg = "failed to create downstream port. Exception: %s" % (str(e))
            LOG.exception(msg)
            raise

    def _get_base_network_info(self, context, network_id, base_net_driver):
        """Return a dict of extra network information.

        :param context: neutron request context.
        :param network_id: neturon network id.
        :param net_driver: network driver associated with network_id.
        :raises IronicException: Any unexpected data fetching failures will
            be logged and IronicException raised.

        This driver can attach to networks managed by other drivers. We may
        need some information from these drivers, or otherwise inform
        downstream about the type of network we are attaching to. We can
        make these decisions here.
        """
        driver_name = base_net_driver.get_name()
        net_info = {"network_type": driver_name}
        LOG.debug('_get_base_network_info: %s %s'
                  % (driver_name, network_id))

        # If the driver is NVP, we need to look up the lswitch id we should
        # be attaching to.
        if driver_name == 'NVP':
            LOG.debug('looking up lswitch ids for network %s'
                      % (network_id))
            lswitch_ids = base_net_driver.get_lswitch_ids_for_network(
                context, network_id)

            if not lswitch_ids or len(lswitch_ids) > 1:
                msg = ('lswitch id lookup failed, %s ids found.'
                       % (len(lswitch_ids)))
                LOG.error(msg)
                raise IronicException(msg)

            lswitch_id = lswitch_ids.pop()
            LOG.info('found lswitch for network %s: %s'
                     % (network_id, lswitch_id))
            net_info['lswitch_id'] = lswitch_id

        LOG.debug('_get_base_network_info finished: %s %s %s'
                  % (driver_name, network_id, net_info))
        return net_info

    def create_port(self, context, network_id, port_id, **kwargs):
        """Create a port.

        :param context: neutron api request context.
        :param network_id: neutron network id.
        :param port_id: neutron port id.
        :param kwargs:
            required keys - device_id: neutron port device_id (instance_id)
                            instance_node_id: nova hypervisor host id
                            mac_address: neutron port mac address
                            base_net_driver: the base network driver
            optional keys - addresses: list of allocated IPAddress models
                            security_groups: list of associated security groups
        :raises IronicException: If the client is unable to create the
            downstream port for any reason, the exception will be logged
            and IronicException raised.
        """
        LOG.info("create_port %s %s %s" % (context.tenant_id, network_id,
                                           port_id))

        # sanity check
        if not kwargs.get('base_net_driver'):
            raise IronicException(msg='base_net_driver required.')
        base_net_driver = kwargs['base_net_driver']

        if not kwargs.get('device_id'):
            raise IronicException(msg='device_id required.')
        device_id = kwargs['device_id']

        if not kwargs.get('instance_node_id'):
            raise IronicException(msg='instance_node_id required.')
        instance_node_id = kwargs['instance_node_id']

        if not kwargs.get('mac_address'):
            raise IronicException(msg='mac_address is required.')
        mac_address = str(netaddr.EUI(kwargs["mac_address"]["address"]))
        mac_address = mac_address.replace('-', ':')

        # TODO(morgabra): Change this when we enable security groups.
        if kwargs.get('security_groups'):
            msg = 'ironic driver does not support security group operations.'
            raise IronicException(msg=msg)

        # unroll the given address models into a fixed_ips list we can
        # pass downstream
        fixed_ips = []
        addresses = kwargs.get('addresses')
        if not isinstance(addresses, list):
            addresses = [addresses]
        for address in addresses:
            fixed_ips.append(self._make_fixed_ip_dict(context, address))

        body = {
            "id": port_id,
            "network_id": network_id,
            "device_id": device_id,
            "device_owner": kwargs.get('device_owner', ''),
            "tenant_id": context.tenant_id or "quark",
            "roles": context.roles,
            "mac_address": mac_address,
            "fixed_ips": fixed_ips,
            "switch:hardware_id": instance_node_id,
            "dynamic_network": not STRATEGY.is_provider_network(network_id)
        }

        net_info = self._get_base_network_info(
            context, network_id, base_net_driver)
        body.update(net_info)

        try:
            LOG.info("creating downstream port: %s" % (body))
            port = self._create_port(context, body)
            LOG.info("created downstream port: %s" % (port))
            return {"uuid": port['port']['id'],
                    "vlan_id": port['port']['vlan_id']}
        except Exception as e:
            msg = "failed to create downstream port. Exception: %s" % (e)
            raise IronicException(msg=msg)

    def update_port(self, context, port_id, **kwargs):
        """Update a port.

        :param context: neutron api request context.
        :param port_id: neutron port id.
        :param kwargs: optional kwargs.
        :raises IronicException: If the client is unable to update the
            downstream port for any reason, the exception will be logged
            and IronicException raised.

        TODO(morgabra) It does not really make sense in the context of Ironic
        to allow updating ports. fixed_ips and mac_address are burned in the
        configdrive on the host, and we otherwise cannot migrate a port between
        instances. Eventually we will need to support security groups, but for
        now it's a no-op on port data changes, and we need to rely on the
        API/Nova to not allow updating data on active ports.
        """
        LOG.info("update_port %s %s" % (context.tenant_id, port_id))

        # TODO(morgabra): Change this when we enable security groups.
        if kwargs.get("security_groups"):
            msg = 'ironic driver does not support security group operations.'
            raise IronicException(msg=msg)

        return {"uuid": port_id}

    @utils.retry_loop(CONF.IRONIC.operation_retries,
                      delay=CONF.IRONIC.operation_delay,
                      backoff=CONF.IRONIC.operation_backoff)
    def _delete_port(self, context, port_id):
        try:
            return self._client.delete_port(port_id)
        except Exception as e:
            # This doesn't get wrapped by the client unfortunately.
            if "404 not found" in str(e).lower():
                LOG.error("port %s not found downstream. ignoring delete."
                          % (port_id))
                return

            msg = ("failed to delete downstream port. "
                   "exception: %s" % (e))
            LOG.exception(msg)
            raise

    def delete_port(self, context, port_id, **kwargs):
        """Delete a port.

        :param context: neutron api request context.
        :param port_id: neutron port id.
        :param kwargs: optional kwargs.
        :raises IronicException: If the client is unable to delete the
            downstream port for any reason, the exception will be logged
            and IronicException raised.
        """
        LOG.info("delete_port %s %s" % (context.tenant_id, port_id))
        try:
            self._delete_port(context, port_id)
            LOG.info("deleted downstream port: %s" % (port_id))
        except Exception:
            LOG.error("failed deleting downstream port, it is now "
                      "orphaned! port_id: %s" % (port_id))

    def diag_port(self, context, port_id, **kwargs):
        """Diagnose a port.

        :param context: neutron api request context.
        :param port_id: neutron port id.
        :param kwargs: optional kwargs.
        :raises IronicException: If the client is unable to fetch the
            downstream port for any reason, the exception will be
            logged and IronicException raised.
        """
        LOG.info("diag_port %s" % port_id)
        try:
            port = self._client.show_port(port_id)
        except Exception as e:
            msg = "failed fetching downstream port: %s" % (str(e))
            LOG.exception(msg)
            raise IronicException(msg=msg)
        return {"downstream_port": port}

    def create_network(self, *args, **kwargs):
        """Create a network.

        :raises NotImplementedError: This driver does not manage networks.

        NOTE: This is a no-op in the base driver, but this raises here as to
        explicitly disallow network operations in case of a misconfiguration.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'network operations.')

    def delete_network(self, *args, **kwargs):
        """Delete a network.

        :raises NotImplementedError: This driver does not manage networks.

        NOTE: This is a no-op in the base driver, but this raises here as to
        explicitly disallow network operations in case of a misconfiguration.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'network operations.')

    def diag_network(self, *args, **kwargs):
        """Diagnose a network.

        :raises NotImplementedError: This driver does not manage networks.

        NOTE: This is a no-op in the base driver, but this raises here as to
        explicitly disallow network operations in case of a misconfiguration.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'network operations.')

    def create_security_group(self, context, group_name, **group):
        """Create a security group.

        :raises NotImplementedError: This driver does not implement security
                                     groups.

        NOTE: Security groups will be supported in the future, but for now
        they are explicitly disallowed.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'security group operations.')

    def delete_security_group(self, context, group_id, **kwargs):
        """Delete a security group.

        :raises NotImplementedError: This driver does not implement security
                                     groups.

        NOTE: Security groups will be supported in the future, but for now
        they are explicitly disallowed.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'security group operations.')

    def update_security_group(self, context, group_id, **group):
        """Update a security group.

        :raises NotImplementedError: This driver does not implement security
                                     groups.

        NOTE: Security groups will be supported in the future, but for now
        they are explicitly disallowed.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'security group operations.')

    def create_security_group_rule(self, context, group_id, rule):
        """Create a security group rule.

        :raises NotImplementedError: This driver does not implement security
                                     groups.

        NOTE: Security groups will be supported in the future, but for now
        they are explicitly disallowed.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'security group operations.')

    def delete_security_group_rule(self, context, group_id, rule):
        """Delete a security group rule.

        :raises NotImplementedError: This driver does not implement security
                                     groups.

        NOTE: Security groups will be supported in the future, but for now
        they are explicitly disallowed.
        """
        raise NotImplementedError('ironic driver does not support '
                                  'security group operations.')



"""indexapalooza

Revision ID: 79d90e62d88
Revises: 33e9e23ba761
Create Date: 2015-05-20 20:52:01.909161

"""

# revision identifiers, used by Alembic.
revision = '79d90e62d88'
down_revision = '33e9e23ba761'

from alembic import op


def upgrade():
    op.drop_index(op.f('ix_quark_ip_addresses_version'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_ip_addresses_deallocated_at'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_ip_addresses_address'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_subnets_tenant_id'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_subnets_segment_id'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_ports_name'), table_name='quark_ports')
    op.drop_index(op.f('ix_quark_ports_device_id'), table_name='quark_ports')
    op.drop_index('idx_ports_3', table_name='quark_ports')
    op.drop_index('idx_ports_2', table_name='quark_ports')
    op.drop_index('idx_ports_1', table_name='quark_ports')
    op.drop_index(op.f('ix_quark_networks_tenant_id'),
                  table_name='quark_networks')
    op.drop_index(op.f('ix_quark_mac_addresses_deallocated_at'),
                  table_name='quark_mac_addresses')
    op.drop_index(op.f('ix_quark_security_groups_tenant_id'),
                  table_name='quark_security_groups')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitchport_port_id'),
                  table_name='quark_nvp_driver_lswitchport')
    op.drop_index(op.f('ix_quark_nvp_driver_security_profile_nvp_id'),
                  table_name='quark_nvp_driver_security_profile')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitch_nvp_id'),
                  table_name='quark_nvp_driver_lswitch')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitch_network_id'),
                  table_name='quark_nvp_driver_lswitch')
    op.drop_index(op.f('ix_quotas_tenant_id'), table_name='quotas')
    op.drop_index(op.f('ix_quark_nvp_orphaned_lswitches_nvp_id'),
                  table_name='quark_nvp_orphaned_lswitches')
    op.drop_index(op.f('ix_quark_nvp_orphaned_lswitches_network_id'),
                  table_name='quark_nvp_orphaned_lswitches')
    op.drop_index(op.f('ix_quark_nvp_orphaned_lswitches_display_name'),
                  table_name='quark_nvp_orphaned_lswitches')
    op.drop_index(op.f('ix_quark_nvp_orphaned_lswitch_ports_port_id'),
                  table_name='quark_nvp_orphaned_lswitch_ports')
    op.drop_index(op.f('ix_quark_ports_created_at'),
                  table_name='quark_ports')
    op.drop_index(op.f('ix_quark_mac_addresses_deallocated'),
                  table_name='quark_mac_addresses')


def downgrade():
    op.create_index(op.f('ix_quark_security_groups_tenant_id'),
                    'quark_security_groups',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_mac_addresses_deallocated_at'),
                    'quark_mac_addresses',
                    ['deallocated_at'],
                    unique=False)
    op.create_index(op.f('ix_quark_networks_tenant_id'),
                    'quark_networks',
                    ['tenant_id'],
                    unique=False)
    op.create_index('idx_ports_1',
                    'quark_ports',
                    ['device_id', 'tenant_id'],
                    unique=False)
    op.create_index('idx_ports_2',
                    'quark_ports',
                    ['device_owner', 'network_id'],
                    unique=False)
    op.create_index('idx_ports_3', 'quark_ports', ['tenant_id'], unique=False)
    op.create_index(op.f('ix_quark_ports_device_id'),
                    'quark_ports',
                    ['device_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_name'), 'quark_ports', ['name'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_segment_id'),
                    'quark_subnets',
                    ['segment_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_tenant_id'),
                    'quark_subnets',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_tenant_id'),
                    'quark_subnets',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ip_addresses_address'),
                    'quark_ip_addresses',
                    ['address'],
                    unique=False)
    op.create_index(op.f('ix_quark_ip_addresses_deallocated_at'),
                    'quark_ip_addresses',
                    ['deallocated_at'],
                    unique=False)
    op.create_index(op.f('ix_quark_ip_addresses_version'),
                    'quark_ip_addresses',
                    ['version'],
                    unique=False)
    op.create_index(op.f('ix_quark_ip_addresses_version'),
                    'quark_ip_addresses',
                    ['version'],
                    unique=False)
    op.create_index(op.f('ix_quotas_tenant_id'),
                    'quotas',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_driver_lswitch_network_id'),
                    'quark_nvp_driver_lswitch',
                    ['network_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_driver_lswitch_nvp_id'),
                    'quark_nvp_driver_lswitch',
                    ['nvp_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_driver_security_profile_nvp_id'),
                    'quark_nvp_driver_security_profile',
                    ['nvp_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_driver_security_profile_nvp_id'),
                    'quark_nvp_driver_security_profile',
                    ['nvp_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_driver_lswitchport_port_id'),
                    'quark_nvp_driver_lswitchport',
                    ['port_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_orphaned_lswitches_nvp_id'),
                    'quark_nvp_orphaned_lswitches',
                    ['nvp_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_orphaned_lswitches_network_id'),
                    'quark_nvp_orphaned_lswitches',
                    ['network_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_orphaned_lswitches_display_name'),
                    'quark_nvp_orphaned_lswitches',
                    ['display_name'],
                    unique=False)
    op.create_index(op.f('ix_quark_nvp_orphaned_lswitch_ports_port_id'),
                    'quark_nvp_orphaned_lswitch_ports',
                    ['port_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_created_at'),
                    'quark_ports',
                    ['created_at'],
                    unique=False)
    op.create_index(op.f('ix_quark_mac_addresses_deallocated'),
                    'quark_mac_addresses',
                    ['deallocated'],
                    unique=False)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from quark.drivers.registry import DriverRegistryBase
from quark.drivers import unicorn_driver as unicorn

CONF = cfg.CONF

quark_router_opts = [
    cfg.StrOpt('default_floating_ip_driver',
               default='Unicorn',
               help=_('Driver for floating IP')),
]

CONF.register_opts(quark_router_opts, 'QUARK')


class FloatingIPDriverRegistry(DriverRegistryBase):
    def __init__(self):
        self.drivers = {
            unicorn.UnicornDriver.get_name(): unicorn.UnicornDriver()}

    def get_driver(self, driver_name=None):
        if not driver_name:
            driver_name = CONF.QUARK.default_floating_ip_driver

        if driver_name in self.drivers:
            return self.drivers[driver_name]

        raise Exception("Driver %s is not registered." % driver_name)

DRIVER_REGISTRY = FloatingIPDriverRegistry()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

quark_opts = [
    cfg.StrOpt('default_net_strategy', default='{}',
               help=_("Default network assignment strategy")),
    cfg.StrOpt('public_net_id', default='00000000-0000-0000-0000-000000000000',
               help=_("Public network id"))
]
CONF.register_opts(quark_opts, "QUARK")


class JSONStrategy(object):
    def __init__(self, strategy=None):
        self.subnet_strategy = {}
        self.strategy = {}
        self.load(strategy)

    def load(self, strategy=None):
        if not strategy:
            self._compile_strategy(CONF.QUARK.default_net_strategy)
        else:
            self._compile_strategy(strategy)

    def _compile_strategy(self, strategy):
        self.strategy = json.loads(strategy)
        for net_id, meta in self.strategy.iteritems():
            if "subnets" in meta:
                for ip_version, subnet_id in meta["subnets"].iteritems():
                    self.subnet_strategy[subnet_id] = {
                        "ip_version": ip_version,
                        "network_id": net_id}
            else:
                LOG.warning('net_id {} strategy has no "subnets" '
                            'metadata.'.format(net_id))

    def _split(self, func, resource_ids):
        provider = []
        tenant = []
        for res_id in resource_ids:
            if func(res_id):
                provider.append(res_id)
            else:
                tenant.append(res_id)
        return tenant, provider

    def split_network_ids(self, net_ids):
        return self._split(self.is_provider_network, net_ids)

    def split_subnet_ids(self, subnet_ids):
        return self._split(self.is_provider_subnet, subnet_ids)

    def get_provider_networks(self):
        return sorted(self.strategy.keys())

    def get_provider_subnets(self):
        return sorted(self.subnet_strategy.keys())

    def get_provider_subnet_id(self, net_id, ip_version):
        if net_id not in self.strategy:
            return None
        return self.strategy[net_id]["subnets"][str(ip_version)]

    def get_network(self, net_id):
        return self.strategy.get(net_id)

    def is_provider_network(self, net_id):
        return self.strategy.get(net_id) is not None

    def is_provider_subnet(self, subnet_id):
        return subnet_id in self.subnet_strategy

    def subnet_ids_for_network(self, net_id):
        if net_id in self.strategy:
            subnets = self.strategy.get(net_id)["subnets"]
            return [subnet_id for ip_version, subnet_id in subnets.iteritems()]

    def get_network_for_subnet(self, subnet_id):
        if subnet_id not in self.subnet_strategy:
            return None
        return self.subnet_strategy.get(subnet_id)["network_id"]

    def get_public_net_id(self):
        """Returns the public net id"""
        for id, net_params in self.strategy.iteritems():
            if id == CONF.QUARK.public_net_id:
                return id
        return None


STRATEGY = JSONStrategy()



from quark.db import api as quark_db_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class QuarkTransactionFunctionalTest(MySqlBaseFunctionalTest):
    def test_transaction_id(self):
        with self.context.session.begin():
            transaction = quark_db_api.transaction_create(self.context)
        self.assertEqual(transaction.id, 1)
        with self.context.session.begin():
            transaction = quark_db_api.transaction_create(self.context)
        self.assertEqual(transaction.id, 2)
        with self.context.session.begin():
            transaction = quark_db_api.transaction_create(self.context)
        self.assertEqual(transaction.id, 3)



"""add scaling ip address type enum

Revision ID: 3f0c11478a5d
Revises: a0798b3b7418
Create Date: 2016-01-22 23:41:03.214930

"""

# revision identifiers, used by Alembic.
revision = '3f0c11478a5d'
down_revision = 'a0798b3b7418'

from alembic import op
import sqlalchemy as sa


existing_enum = sa.Enum("shared", "floating", "fixed")
new_enum = sa.Enum("shared", "floating", "fixed", "scaling")


def upgrade():
    op.alter_column("quark_ip_addresses", "address_type",
                    existing_type=existing_enum,
                    type_=new_enum)


def downgrade():
    op.alter_column("quark_ip_addresses", "address_type",
                    existing_type=new_enum,
                    type_=existing_enum)



"""populate address_type column

Revision ID: 4fc07b41d45c
Revises: 42a3c8c0db75
Create Date: 2014-10-16 14:09:43.175030

"""

# revision identifiers, used by Alembic.
revision = '4fc07b41d45c'
down_revision = '42a3c8c0db75'

from alembic import op
from sqlalchemy.sql import table, column
import sqlalchemy as sa


def upgrade():
    ip_addresses = table('quark_ip_addresses',
                         column('address_type', sa.Enum),
                         column('_deallocated', sa.Boolean))
    connection = op.get_bind()
    t = ip_addresses.update().values({'address_type': 'fixed'}).where(
        ip_addresses.c._deallocated == 0)
    connection.execute(t)


def downgrade():
    ip_addresses = table('quark_ip_addresses',
                         column('address_type', sa.Enum))
    connection = op.get_bind()
    t = ip_addresses.update().values({'address_type': None})
    connection.execute(t)



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from quark.tests import test_base
from quark import utils


def func_do(**kwargs):
    if 'rollback' in kwargs and kwargs['rollback']:
        return True
    else:
        return False


def func_undo(**kwargs):
    if 'rollback' in kwargs and kwargs['rollback']:
        return True
    else:
        return False


class QuarkCommandManagerTest(test_base.TestBase):
    def setUp(self):
        super(QuarkCommandManagerTest, self).setUp()
        # Using globals here because of callbacks
        self.is_rollback_do = False
        self.is_rollback_undo = True

    @mock.patch('quark.tests.test_command_manager.func_undo')
    @mock.patch('quark.tests.test_command_manager.func_do')
    def test_command_manager_no_undo(self,
                                     func_do_notifier,
                                     func_undo_notifier):
        """Test that undo is not called when everything is good"""
        try:
            with utils.CommandManager().execute() as cmd_mgr:
                @cmd_mgr.do
                def f(**kwargs):
                    func_do(**kwargs)

                @cmd_mgr.undo
                def f_undo(*args, **kwargs):
                    func_undo(**kwargs)

                f()
        except Exception:
            pass

        self.assertTrue(func_do_notifier.called)
        self.assertFalse(func_undo_notifier.called)

    @mock.patch('quark.tests.test_command_manager.func_undo')
    def test_command_manager_undo(self, func_undo_notifier):
        """Test that undo is called when the do function raises"""
        try:
            with utils.CommandManager().execute() as cmd_mgr:
                @cmd_mgr.do
                def f(**kwargs):
                    raise Exception

                @cmd_mgr.undo
                def f_undo(*args, **kwargs):
                    func_undo(**kwargs)

                f()
        except Exception:
            pass

        self.assertTrue(func_undo_notifier.called)

    def test_rollback_is_passed_to_do(self):
        """Tests that the do function has rollback set to False"""
        self.is_rollback_do = True
        try:
            with utils.CommandManager().execute() as cmd_mgr:
                @cmd_mgr.do
                def f(**kwargs):
                    return func_do(**kwargs)

                @cmd_mgr.undo
                def f_undo(*args, **kwargs):
                    func_undo(**kwargs)

                self.is_rollback_do = f()

        except Exception:
            pass

        self.assertFalse(self.is_rollback_do)

    def test_rollback_is_passed_to_undo(self):
        """Tests that the undo function has rollback set to True"""
        self.is_rollback_undo = False
        try:
            with utils.CommandManager().execute() as cmd_mgr:
                @cmd_mgr.do
                def f(**kwargs):
                    raise Exception

                @cmd_mgr.undo
                def f_undo(*args, **kwargs):
                    self.is_rollback_undo = func_undo(**kwargs)

                f()
        except Exception:
            pass

        self.assertTrue(self.is_rollback_undo)



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

import functools
import json
import string

import netaddr
from oslo_config import cfg
from oslo_log import log as logging

from quark import exceptions as q_exc

from twiceredis import TwiceRedis


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
MAC_TRANS_TABLE = string.maketrans(string.ascii_uppercase,
                                   string.ascii_lowercase)

quark_opts = [
    cfg.ListOpt("redis_sentinel_hosts",
                default=["localhost:26397"],
                help=_("Comma-separated list of host:port pairs for Redis "
                       "sentinel hosts")),
    cfg.StrOpt("redis_sentinel_master",
               default='',
               help=_("The name label of the master redis sentinel")),
    cfg.StrOpt("redis_password",
               default='',
               help=_("The password for authenticating with redis.")),
    cfg.StrOpt("redis_db",
               default="0",
               help=("The database number to use")),
    cfg.FloatOpt("redis_socket_timeout",
                 default=0.1,
                 help=("Timeout for Redis socket operations"))]

CONF.register_opts(quark_opts, "QUARK")

# TODO(mdietz): Rewrite this to use a module level connection
#               pool, and then incorporate that into creating
#               connections. When connecting to a master we
#               connect by creating a redis client, and when
#               we connect to a slave, we connect by telling it
#               we want a slave and ending up with a connection,
#               with no control over SSL or anything else.  :-|


def handle_connection_error(fn):
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except TwiceRedis.generic_error as e:
            LOG.exception(e)
            raise q_exc.RedisConnectionFailure()
    return wrapped


class ClientBase(object):
    def __init__(self):
        self._client = self.get_redis_client()

    def get_redis_client(self):
        sentinels = [tuple(str.split(host_pair, ':'))
                     for host_pair in CONF.QUARK.redis_sentinel_hosts]

        pool_kwargs = TwiceRedis.DEFAULT_POOL_KWARGS
        sentinel_kwargs = TwiceRedis.DEFAULT_SENTINEL_KWARGS
        pool_kwargs['socket_timeout'] = CONF.QUARK.redis_socket_timeout
        pool_kwargs['socket_keepalive'] = False
        sentinel_kwargs['min_other_sentinels'] = 2
        return TwiceRedis(master_name=CONF.QUARK.redis_sentinel_master,
                          sentinels=sentinels,
                          password=CONF.QUARK.redis_password,
                          pool_kwargs=pool_kwargs,
                          sentinel_kwargs=sentinel_kwargs)

    def vif_key(self, device_id, mac_address):
        mac = str(netaddr.EUI(mac_address))

        # Lower cases and strips hyphens from the mac
        mac = mac.translate(MAC_TRANS_TABLE, ":-")
        return "{0}.{1}".format(device_id, mac)

    @handle_connection_error
    def ping(self):
        # NOTE(tr3buchet): if this gets used by anything other than the
        #                  redis_sg_tool, self._client.disconnect()
        #                  needs to be called before returning
        return self._client.master.ping() and self._client.slave.ping()

    @handle_connection_error
    def vif_keys(self, field=None):
        keys = self._client.slave.keys("*.????????????")
        filtered = []
        if isinstance(keys, str):
            keys = [keys]
        with self._client.slave.pipeline() as pipe:
            for key in keys:
                value = None
                if field:
                    value = pipe.hget(key, field)
                else:
                    value = pipe.hgetall(key)
            values = pipe.execute()
        for value in values:
            if value:
                filtered.append(key)
        return filtered

    @handle_connection_error
    def set_field(self, key, field, data):
        self.set_field_raw(key, field, json.dumps(data))

    @handle_connection_error
    def set_field_raw(self, key, field, data):
        self._client.master.hset(key, field, data)
        self._client.master.disconnect()

    @handle_connection_error
    def get_field(self, key, field):
        return self._client.slave.hget(key, field)

    @handle_connection_error
    def delete_field(self, key, field):
        self._client.master.hdel(key, field)
        self._client.master.disconnect()

    @handle_connection_error
    def delete_key(self, key):
        self._client.master.delete(key)
        self._client.master.disconnect()

    @handle_connection_error
    def get_fields(self, keys, field):
        with self._client.slave.pipeline() as pipe:
            for key in keys:
                pipe.hget(key, field)
            values = pipe.execute()
        return values

    @handle_connection_error
    def set_fields(self, keys, field, value):
        with self._client.master.pipeline() as pipe:
            for key in keys:
                pipe.hset(key, field, value)
            pipe.execute()
        self._client.master.disconnect()



"""composite primary key port ip association

Revision ID: 3b467be51e43
Revises: 26e984b48a0d
Create Date: 2014-10-09 15:42:58.104964

"""

# revision identifiers, used by Alembic.
revision = '3b467be51e43'
down_revision = '26e984b48a0d'

import contextlib
import itertools

from alembic import op
import sqlalchemy as sa


def _drop_foreign_key(op, table, fk):
    op.drop_constraint(fk.name, table.name, type_="foreignkey")


def _create_foreign_key(op, table, fk):
    ref_tbl_name = fk.column.table.name
    ref_tbl_col = fk.column.name
    op.create_foreign_key(fk.name, table.name, ref_tbl_name,
                          [fk.parent.name], [ref_tbl_col])


def _alter_foreign_keys(op, table, action, fk_constraints):
    actions = {
        "drop": _drop_foreign_key,
        "create": _create_foreign_key
    }
    # NOTE(thomasem): Flatten list of ForeignKeys we want to work on from the
    # ForeignKeyConstraint objects that may contain multiple ForeignKey
    # objects.
    [actions[action](op, table, fk) for fk
     in itertools.chain.from_iterable([c.elements for c in fk_constraints])]


@contextlib.contextmanager
def _foreign_keys_dropped(op, table):
    fk_constraints = [c for c in table.constraints
                      if isinstance(c, sa.schema.ForeignKeyConstraint)]
    _alter_foreign_keys(op, table, "drop", fk_constraints)
    yield
    _alter_foreign_keys(op, table, "create", fk_constraints)


def upgrade():
    metadata = sa.MetaData(bind=op.get_bind())
    table = sa.Table('quark_port_ip_address_associations', metadata,
                     autoload=True)
    with _foreign_keys_dropped(op, table):
        op.alter_column('quark_port_ip_address_associations', 'ip_address_id',
                        existing_type=sa.String(36), nullable=False)
        op.alter_column('quark_port_ip_address_associations', 'port_id',
                        existing_type=sa.String(36), nullable=False)
        op.create_primary_key("pk_quark_port_ip_address_associations",
                              "quark_port_ip_address_associations",
                              ['port_id', 'ip_address_id'])


def downgrade():
    metadata = sa.MetaData(bind=op.get_bind())
    table = sa.Table('quark_port_ip_address_associations', metadata,
                     autoload=True)
    # NOTE(thomasem): Unfortunately we cannot remove primary keys for columns
    # that have a ForeignKeyConstraint defined. So, we can temporarily remove
    # them and add them back as soon as the PrimaryKeyConstraint is removed.
    with _foreign_keys_dropped(op, table):
        op.drop_constraint("pk_quark_port_ip_address_associations",
                           "quark_port_ip_address_associations",
                           type_="primary")
        op.alter_column('quark_port_ip_address_associations', 'port_id',
                        existing_type=sa.String(36), nullable=True)
        op.alter_column('quark_port_ip_address_associations', 'ip_address_id',
                        existing_type=sa.String(36), nullable=True)



"""Populate first_ip, last_ip for each IP Policy CIDR

Revision ID: 1664300cb03a
Revises: 1acd075bd7e1
Create Date: 2014-08-07 07:17:32.841155

"""

# revision identifiers, used by Alembic.
revision = '1664300cb03a'
down_revision = '1acd075bd7e1'

from alembic import op
import netaddr
from sqlalchemy.sql import column, select, table
import sqlalchemy as sa

from quark.db.custom_types import INET


def upgrade():
    ip_policy_cidrs = table('quark_ip_policy_cidrs',
                            column('id', sa.String(length=36)),
                            column('first_ip', INET()),
                            column('last_ip', INET()),
                            column('cidr', sa.String(length=64)))
    connection = op.get_bind()

    # 1. Retrieve all ip_policy_cidr rows.
    results = connection.execute(
        select([ip_policy_cidrs.c.id, ip_policy_cidrs.c.cidr])
    ).fetchall()

    # 2. Populate first_ip, last_ip for each IP Policy CIDR.
    for ippc in results:
        net = netaddr.IPNetwork(ippc["cidr"]).ipv6()
        connection.execute(ip_policy_cidrs.update().values(
            first_ip=net.first, last_ip=net.last).where(
                ip_policy_cidrs.c.id == ippc["id"]))


def downgrade():
    raise NotImplementedError()



# Copyright 2016 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import plugin_views as v

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def get_jobs(context, **filters):
    LOG.info("get_jobs for tenant %s" % context.tenant_id)
    if not filters:
        filters = {}
    jobs = db_api.async_transaction_find(context, scope=db_api.ALL, **filters)
    return [v._make_job_dict(ip) for ip in jobs]


def get_job(context, id):
    LOG.info("get_job %s for tenant %s" % (id, context.tenant_id))
    filters = {}
    job = db_api.async_transaction_find(context, id=id, scope=db_api.ONE,
                                        **filters)
    if not job:
        raise q_exc.JobNotFound(job_id=id)
    return v._make_job_dict(job)


def create_job(context, body):
    LOG.info("create_job for tenant %s" % context.tenant_id)

    if not context.is_admin:
        raise n_exc.NotAuthorized()
    job = body.get('job')
    if not job:
        raise n_exc.BadRequest(resource="job", msg="Invalid request body.")
    with context.session.begin():
        new_job = db_api.async_transaction_create(context, **job)
    return v._make_job_dict(new_job)


def update_job(context, id, body):
    LOG.info("update_job %s for tenant %s" % (id, context.tenant_id))

    if not context.is_admin:
        raise n_exc.NotAuthorized()
    job_update = body.get('job')
    if not job_update:
        raise n_exc.BadRequest(resource="job", msg="Invalid request body.")
    job = db_api.async_transaction_find(context, id=id, scope=db_api.ONE)
    if not job:
        raise q_exc.JobNotFound(job_id=id)
    job_mod = db_api.async_transaction_update(context, job, **job_update)
    return v._make_job_dict(job_mod)


def delete_job(context, id, **filters):
    """Delete an ip address.

    : param context: neutron api request context
    : param id: UUID representing the ip address to delete.
    """
    LOG.info("delete_ip_address %s for tenant %s" % (id, context.tenant_id))

    if not context.is_admin:
        raise n_exc.NotAuthorized()
    with context.session.begin():
        job = db_api.async_transaction_find(context, id=id, scope=db_api.ONE,
                                            **filters)
        if not job:
            raise q_exc.JobNotFound(job_id=id)
        db_api.async_transaction_delete(context, job)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
View Helpers for Quark Plugin
"""

import netaddr
from oslo_config import cfg
from oslo_log import log as logging

from quark.db import ip_types
from quark import network_strategy
from quark import protocols
from quark import tags


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
STRATEGY = network_strategy.STRATEGY
PORT_TAG_REGISTRY = tags.PORT_TAG_REGISTRY

quark_view_opts = [
    cfg.BoolOpt('show_allocation_pools',
                default=True,
                help=_('Controls whether or not to calculate and display'
                       'allocation pools or not')),
    cfg.BoolOpt('show_ipam_strategy',
                default=False,
                help=_('Controls whether or not to show ipam_strategy')),
    cfg.BoolOpt('show_subnet_ip_policy_id',
                default=True,
                help=_('Controls whether or not to show ip_policy_id for'
                       'subnets')),
    cfg.BoolOpt('show_provider_subnet_ids',
                default=True,
                help=_('Controls whether or not to show the provider subnet '
                       'id specified in the network strategy or use the '
                       'real id.')),
]

CONF.register_opts(quark_view_opts, "QUARK")


def _is_default_route(route):
    return route.value == 0


def _make_network_dict(network, fields=None):
    shared_net = STRATEGY.is_provider_network(network["id"])
    res = {"id": network["id"],
           "name": network.get("name"),
           "tenant_id": network.get("tenant_id"),
           "admin_state_up": True,
           "status": "ACTIVE",
           "shared": shared_net}
    if CONF.QUARK.show_ipam_strategy:
        res['ipam_strategy'] = network.get("ipam_strategy")

    if not shared_net:
        if fields and "all_subnets" in fields:
            res["subnets"] = [_make_subnet_dict(s)
                              for s in network.get("subnets", [])]
        else:
            res["subnets"] = [s["id"] for s in network.get("subnets", [])]
    else:
        res["subnets"] = STRATEGY.subnet_ids_for_network(network["id"])
    return res


def _make_subnet_dict(subnet, fields=None):
    dns_nameservers = [str(netaddr.IPAddress(dns["ip"]))
                       for dns in subnet.get("dns_nameservers", [])]
    subnet_id = subnet.get("id")
    if STRATEGY.is_provider_subnet(subnet_id):
        net_id = STRATEGY.get_network_for_subnet(subnet_id)
    else:
        net_id = subnet["network_id"]

    res = {"id": subnet_id,
           "name": subnet.get("name"),
           "tenant_id": subnet.get("tenant_id"),
           "network_id": net_id,
           "ip_version": subnet.get("ip_version"),
           "dns_nameservers": dns_nameservers or [],
           "cidr": subnet.get("cidr"),
           "shared": STRATEGY.is_provider_network(net_id),
           "enable_dhcp": None}

    if CONF.QUARK.show_subnet_ip_policy_id:
        res['ip_policy_id'] = subnet.get("ip_policy_id")

    if (CONF.QUARK.show_allocation_pools and not
            STRATEGY.is_provider_subnet(subnet_id)):
        res["allocation_pools"] = subnet.get('allocation_pools', [])
    else:
        res["allocation_pools"] = []

    def _host_route(route):
        return {"destination": route["cidr"],
                "nexthop": route["gateway"]}

    res["gateway_ip"] = None
    res["host_routes"] = []
    default_found = False
    for route in subnet.get("routes", []):
        netroute = netaddr.IPNetwork(route["cidr"])
        if _is_default_route(netroute):
            # NOTE(mdietz): This has the potential to find more than one
            #       default route. Quark normally won't allow you to create
            #       more than one, but it's plausible one exists regardless.
            #       As such, we're going to pretend it isn't possible, but
            #       log it anyway.
            if default_found:
                LOG.info(_("Default route %(gateway_ip)s already found for "
                           "subnet %(id)s") % res)
            res["gateway_ip"] = route["gateway"]
            default_found = True
        else:
            res["host_routes"].append(_host_route(route))
    return res


def _make_security_group_dict(security_group, fields=None):
    res = {"id": security_group.get("id"),
           "description": security_group.get("description"),
           "name": security_group.get("name"),
           "tenant_id": security_group.get("tenant_id")}
    res["security_group_rules"] = [
        _make_security_group_rule_dict(r) for r in security_group["rules"]]
    return res


def _make_security_group_rule_dict(security_rule, fields=None):
    ethertype = protocols.human_readable_ethertype(
        security_rule.get("ethertype"))
    protocol = protocols.human_readable_protocol(
        security_rule.get("protocol"), ethertype)

    res = {"id": security_rule.get("id"),
           "ethertype": ethertype,
           "direction": security_rule.get("direction"),
           "tenant_id": security_rule.get("tenant_id"),
           "port_range_max": security_rule.get("port_range_max"),
           "port_range_min": security_rule.get("port_range_min"),
           "protocol": protocol,
           "remote_ip_prefix": security_rule.get("remote_ip_prefix"),
           "security_group_id": security_rule.get("group_id"),
           "remote_group_id": security_rule.get("remote_group_id")}
    return res


def _ip_port_dict(ip, port, fields=None):
    service = ip.get_service_for_port(port)
    res = {"id": port.get("id"),
           "device_id": port.get("device_id"),
           "service": service}
    return res


def _port_dict(port, fields=None):
    res = {"id": port.get("id"),
           "name": port.get("name"),
           "network_id": port["network_id"],
           "tenant_id": port.get("tenant_id"),
           "mac_address": port.get("mac_address"),
           "admin_state_up": port.get("admin_state_up"),
           "status": "ACTIVE",
           "security_groups": [group.get("id", None) for group in
                               port.get("security_groups", None)],
           "device_id": port.get("device_id"),
           "device_owner": port.get("device_owner")}

    if "mac_address" in res and res["mac_address"]:
        mac = str(netaddr.EUI(res["mac_address"])).replace('-', ':')
        res["mac_address"] = mac

    # NOTE(mdietz): more pythonic key in dict check fails here. Leave as get
    if port.get("bridge"):
        res["bridge"] = port["bridge"]

    # NOTE(ClifHouck): This causes another trip to the DB since tags are
    # are not eager loaded. According to mdietz this be a small impact on
    # performance, but if the tag system gets used more on ports, we may
    # want to eager load the tags.
    try:
        t = PORT_TAG_REGISTRY.get_all(port)
        res.update(t)
    except Exception as e:
        # NOTE(morgabra) We really don't want to break port-listing if
        # this goes sideways here, so we pass.
        msg = ("Unknown error loading tags for port %s: %s"
               % (port["id"], e))
        LOG.exception(msg)

    return res


def _make_port_address_dict(ip, port, fields=None):
    enabled = ip.enabled_for_port(port)
    subnet_id = ip.get("subnet_id")
    net_id = ip.get("network_id")

    show_provider_subnet_ids = CONF.QUARK.show_provider_subnet_ids
    if STRATEGY.is_provider_network(net_id) and show_provider_subnet_ids:
            subnet_id = STRATEGY.get_provider_subnet_id(
                net_id, ip["version"])

    ip_addr = {"subnet_id": subnet_id,
               "ip_address": ip.formatted(),
               "enabled": enabled}
    if fields and "port_subnets" in fields:
        ip_addr["subnet"] = _make_subnet_dict(ip["subnet"])

    return ip_addr


def _make_port_for_ip_dict(ip, port, fields=None):
    res = _ip_port_dict(ip, port)
    return res


def _ip_is_fixed(port, ip):
    at = ip.get('address_type')
    return (not at or at in (ip_types.FIXED, ip_types.SHARED))


def _make_port_dict(port, fields=None):
    res = _port_dict(port)
    ips = []
    for assoc in port.associations:
        ips.append(assoc.ip_address)
    res["fixed_ips"] = [_make_port_address_dict(ip, port, fields)
                        for ip in ips if _ip_is_fixed(port, ip)]
    return res


def _make_ip_ports_list(ip, query, fields=None):
    ports = []
    for port in query:
        port_dict = _ip_port_dict(ip, port, fields)
        ports.append(port_dict)
    return ports


def _make_ports_list(query, fields=None):
    ports = []
    for port in query:
        port_dict = _port_dict(port, fields)
        port_dict["fixed_ips"] = [_make_port_address_dict(ip, port, fields)
                                  for ip in port.ip_addresses if
                                  _ip_is_fixed(port, ip)]
        ports.append(port_dict)
    return ports


def _make_subnets_list(query, fields=None):
    subnets = []
    for subnet in query:
        subnets.append(_make_subnet_dict(subnet, fields=fields))
    return subnets


def _make_mac_range_dict(mac_range):
    return {"id": mac_range["id"],
            "cidr": mac_range["cidr"]}


def _make_segment_allocation_range_dict(sa_range, allocations=None):
    size = len(xrange(sa_range["first_id"], sa_range["last_id"] + 1))

    sa_dict = {
        "id": sa_range["id"],
        "segment_id": sa_range["segment_id"],
        "segment_type": sa_range["segment_type"],
        "first_id": sa_range["first_id"],
        "last_id": sa_range["last_id"],
        "do_not_use": sa_range["do_not_use"],
        "size": size}

    if allocations is not None:
        sa_dict["free_ids"] = sa_dict["size"] - allocations
    return sa_dict


def _make_route_dict(route):
    return {"id": route["id"],
            "cidr": route["cidr"],
            "gateway": route["gateway"],
            "subnet_id": route["subnet_id"]}


def _make_ip_dict(address):
    return {"id": address["id"],
            "network_id": address["network_id"],
            "ip_address": address.formatted(),
            "address": address.formatted(),
            "port_ids": [assoc.port_id
                         for assoc in address["associations"]],
            "subnet_id": address["subnet_id"],
            "tenant_id": address["used_by_tenant_id"],
            "version": address["version"],
            "type": address['address_type']}


def _make_ip_policy_dict(ipp):
    return {"id": ipp["id"],
            "tenant_id": ipp["tenant_id"],
            "name": ipp["name"],
            "subnet_ids": [s["id"] for s in ipp["subnets"]],
            "network_ids": [n["id"] for n in ipp["networks"]],
            "exclude": [ippc["cidr"] for ippc in ipp["exclude"]]}


def _make_floating_ip_dict(flip, port_id=None):
    if not port_id:
        ports = flip.ports
        port_id = None
        if ports and len(ports) > 0:
            port_id = None if not ports[0] else ports[0].id

    fixed_ip = flip.fixed_ips[0] if flip.fixed_ips else None

    return {"id": flip.get("id"),
            "floating_network_id": flip.get("network_id"),
            "router_id": CONF.QUARK.floating_ip_router_id,
            "fixed_ip_address": None if not fixed_ip else fixed_ip.formatted(),
            "floating_ip_address": flip.formatted(),
            "tenant_id": flip.get("used_by_tenant_id"),
            "status": "RESERVED" if not port_id else "ASSOCIATED",
            "port_id": port_id}


def _make_scaling_ip_dict(flip):
    # Can an IPAddress.fixed_ip have more than one port associated with it?
    ports = []
    for fixed_ip in flip.fixed_ips:
        if fixed_ip.ports:
            ports.append({"port_id": fixed_ip.ports[0].id,
                          "fixed_ip_address": fixed_ip.address_readable})
    return {"id": flip.get("id"),
            "scaling_ip_address": None if not flip else flip.formatted(),
            "scaling_network_id": flip.get("network_id"),
            "tenant_id": flip.get("used_by_tenant_id"),
            "status": flip.get("status"),
            "ports": ports}


def _make_job_dict(job):
    return {"id": job.get('id'),
            "action": job.get('action'),
            "completed": job.get('completed'),
            "tenant_id": job.get('tenant_id'),
            "created_at": job.get('created_at')}
